from abc import ABC, abstractmethod
from typing import Union, Optional
from iqrfpy.enums.commands import *
from iqrfpy.enums.message_types import *
from iqrfpy.enums.peripherals import *
from iqrfpy.async_response import AsyncResponse
from iqrfpy.confirmation import Confirmation
from iqrfpy.iresponse import IResponse
from iqrfpy.peripherals.generic.responses import GenericResponse
import iqrfpy.peripherals.binaryoutput.responses as bo_responses
import iqrfpy.peripherals.coordinator.responses as c_responses
import iqrfpy.peripherals.eeeprom.responses as eeeprom_responses
import iqrfpy.peripherals.eeprom.responses as eeprom_responses
import iqrfpy.peripherals.exploration.responses as exploration_responses
import iqrfpy.peripherals.frc.responses as frc_responses
import iqrfpy.peripherals.io.responses as io_responses
import iqrfpy.peripherals.ledg.responses as ledg_responses
import iqrfpy.peripherals.ledr.responses as ledr_responses
import iqrfpy.peripherals.node.responses as node_responses
import iqrfpy.peripherals.os.responses as os_responses
import iqrfpy.peripherals.ram.responses as ram_responses
import iqrfpy.peripherals.sensor.responses as sensor_responses
import iqrfpy.peripherals.thermometer.responses as thermometer_responses
import iqrfpy.peripherals.uart.responses as uart_responses
from iqrfpy.utils.common import Common
from iqrfpy.utils.dpa import *
from iqrfpy.exceptions import UnsupportedMessageTypeError, UnsupportedPeripheralError, \
    UnsupportedPeripheralCommandError, ResponseFactoryRegisterError

__all__ = [
    'ResponseFactory',
    'BaseFactory',
]


class BaseFactory(ABC):
    """Response factory class abstraction."""

    @abstractmethod
    def create_from_dpa(self, dpa: bytes) -> IResponse:
        """Returns a response object created from DPA message."""

    @abstractmethod
    def create_from_json(self, json: dict) -> IResponse:
        """Returns a response object created from JSON API message."""


# Generic factories
class GenericResponseFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> GenericResponse:
        return GenericResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> IResponse:
        return GenericResponse.from_json(json=json)


class AsyncResponseFactory(BaseFactory):
    def create_from_dpa(self, dpa: bytes) -> IResponse:
        return AsyncResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> IResponse:
        return AsyncResponse.from_json(json=json)


class ConfirmationFactory(BaseFactory):
    def create_from_dpa(self, dpa: bytes) -> Confirmation:
        return Confirmation.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> Confirmation:
        return Confirmation.from_json(json=json)


# Coordinator factories
class CoordinatorAddrInfoFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> c_responses.AddrInfoResponse:
        return c_responses.AddrInfoResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> c_responses.AddrInfoResponse:
        return c_responses.AddrInfoResponse.from_json(json=json)


class CoordinatorAuthorizeBondFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> c_responses.AuthorizeBondResponse:
        return c_responses.AuthorizeBondResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> c_responses.AuthorizeBondResponse:
        return c_responses.AuthorizeBondResponse.from_json(json=json)


class CoordinatorBackupFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> c_responses.BackupResponse:
        return c_responses.BackupResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> c_responses.BackupResponse:
        return c_responses.BackupResponse.from_json(json=json)


class CoordinatorBondedDevicesFactory(BaseFactory):
    def create_from_dpa(self, dpa: bytes) -> c_responses.BondedDevicesResponse:
        return c_responses.BondedDevicesResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> c_responses.BondedDevicesResponse:
        return c_responses.BondedDevicesResponse.from_json(json=json)


class CoordinatorBondNodeFactory(BaseFactory):
    def create_from_dpa(self, dpa: bytes) -> c_responses.BondNodeResponse:
        return c_responses.BondNodeResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> c_responses.BondNodeResponse:
        return c_responses.BondNodeResponse.from_json(json=json)


class CoordinatorClearAllBondsFactory(BaseFactory):
    def create_from_dpa(self, dpa: bytes) -> c_responses.ClearAllBondsResponse:
        return c_responses.ClearAllBondsResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> c_responses.ClearAllBondsResponse:
        return c_responses.ClearAllBondsResponse.from_json(json=json)


class CoordinatorDiscoveredDevicesFactory(BaseFactory):
    def create_from_dpa(self, dpa: bytes) -> c_responses.DiscoveredDevicesResponse:
        return c_responses.DiscoveredDevicesResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> c_responses.DiscoveredDevicesResponse:
        return c_responses.DiscoveredDevicesResponse.from_json(json=json)


class CoordinatorDiscoveryFactory(BaseFactory):
    def create_from_dpa(self, dpa: bytes) -> c_responses.DiscoveryResponse:
        return c_responses.DiscoveryResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> c_responses.DiscoveryResponse:
        return c_responses.DiscoveryResponse.from_json(json=json)


class CoordinatorRemoveBondFactory(BaseFactory):
    def create_from_dpa(self, dpa: bytes) -> c_responses.RemoveBondResponse:
        return c_responses.RemoveBondResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> c_responses.RemoveBondResponse:
        return c_responses.RemoveBondResponse.from_json(json=json)


class CoordinatorRestoreFactory(BaseFactory):
    def create_from_dpa(self, dpa: bytes) -> c_responses.RestoreResponse:
        return c_responses.RestoreResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> c_responses.RestoreResponse:
        return c_responses.RestoreResponse.from_json(json=json)


class CoordinatorSetDpaParamsFactory(BaseFactory):
    def create_from_dpa(self, dpa: bytes) -> c_responses.SetDpaParamsResponse:
        return c_responses.SetDpaParamsResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> c_responses.SetDpaParamsResponse:
        return c_responses.SetDpaParamsResponse.from_json(json=json)


class CoordinatorSetHopsFactory(BaseFactory):
    def create_from_dpa(self, dpa: bytes) -> c_responses.SetHopsResponse:
        return c_responses.SetHopsResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> c_responses.SetHopsResponse:
        return c_responses.SetHopsResponse.from_json(json=json)


class CoordinatorSetMIDFactory(BaseFactory):
    def create_from_dpa(self, dpa: bytes) -> c_responses.SetMidResponse:
        return c_responses.SetMidResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> c_responses.SetMidResponse:
        return c_responses.SetMidResponse.from_json(json=json)


class CoordinatorSmartConnectFactory(BaseFactory):
    def create_from_dpa(self, dpa: bytes) -> c_responses.SmartConnectResponse:
        return c_responses.SmartConnectResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> c_responses.SmartConnectResponse:
        return c_responses.SmartConnectResponse.from_json(json=json)


# EEEPROM factories
class EeepromReadFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> eeeprom_responses.ReadResponse:
        return eeeprom_responses.ReadResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> eeeprom_responses.ReadResponse:
        return eeeprom_responses.ReadResponse.from_json(json=json)


class EeepromWriteFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> eeeprom_responses.WriteResponse:
        return eeeprom_responses.WriteResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> eeeprom_responses.WriteResponse:
        return eeeprom_responses.WriteResponse.from_json(json=json)


# EEPROM factories
class EepromReadFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> eeprom_responses.ReadResponse:
        return eeprom_responses.ReadResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> eeprom_responses.ReadResponse:
        return eeprom_responses.ReadResponse.from_json(json=json)


class EepromWriteFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> eeprom_responses.WriteResponse:
        return eeprom_responses.WriteResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> eeprom_responses.WriteResponse:
        return eeprom_responses.WriteResponse.from_json(json=json)


# Exploration factories
class ExplorationPeripheralEnumerationFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> exploration_responses.PeripheralEnumerationResponse:
        return exploration_responses.PeripheralEnumerationResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> exploration_responses.PeripheralEnumerationResponse:
        return exploration_responses.PeripheralEnumerationResponse.from_json(json=json)


class ExplorationPeripheralInformationFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> exploration_responses.PeripheralInformationResponse:
        return exploration_responses.PeripheralInformationResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> exploration_responses.PeripheralInformationResponse:
        return exploration_responses.PeripheralInformationResponse.from_json(json=json)


class ExplorationMorePeripheralsInformationFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> exploration_responses.MorePeripheralsInformationResponse:
        return exploration_responses.MorePeripheralsInformationResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> exploration_responses.MorePeripheralsInformationResponse:
        return exploration_responses.MorePeripheralsInformationResponse.from_json(json=json)


# FRC factories
class FrcSendFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> frc_responses.SendResponse:
        return frc_responses.SendResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> frc_responses.SendResponse:
        return frc_responses.SendResponse.from_json(json=json)


class FrcExtraResultFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> frc_responses.ExtraResultResponse:
        return frc_responses.ExtraResultResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> frc_responses.ExtraResultResponse:
        return frc_responses.ExtraResultResponse.from_json(json=json)


class FrcSendSelectiveFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> frc_responses.SendSelectiveResponse:
        return frc_responses.SendSelectiveResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> frc_responses.SendSelectiveResponse:
        return frc_responses.SendSelectiveResponse.from_json(json=json)


class FrcSetFrcParamsFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> frc_responses.SetFrcParamsResponse:
        return frc_responses.SetFrcParamsResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> frc_responses.SetFrcParamsResponse:
        return frc_responses.SetFrcParamsResponse.from_json(json=json)


# IO factories
class IoDirectionFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> io_responses.DirectionResponse:
        return io_responses.DirectionResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> io_responses.DirectionResponse:
        return io_responses.DirectionResponse.from_json(json=json)


class IoGetFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> io_responses.GetResponse:
        return io_responses.GetResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> io_responses.GetResponse:
        return io_responses.GetResponse.from_json(json=json)


class IoSetFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> io_responses.SetResponse:
        return io_responses.SetResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> io_responses.SetResponse:
        return io_responses.SetResponse.from_json(json=json)


# LEDG factories
class LedgSetOnFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> ledg_responses.SetOnResponse:
        return ledg_responses.SetOnResponse.from_dpa(dpa)

    def create_from_json(self, json: dict) -> ledg_responses.SetOnResponse:
        return ledg_responses.SetOnResponse.from_json(json)


class LedgSetOffFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> ledg_responses.SetOffResponse:
        return ledg_responses.SetOffResponse.from_dpa(dpa)

    def create_from_json(self, json: dict) -> ledg_responses.SetOffResponse:
        return ledg_responses.SetOffResponse.from_json(json)


class LedgPulseFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> ledg_responses.PulseResponse:
        return ledg_responses.PulseResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> ledg_responses.PulseResponse:
        return ledg_responses.PulseResponse.from_json(json=json)


class LedgFlashingFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> ledg_responses.FlashingResponse:
        return ledg_responses.FlashingResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> ledg_responses.FlashingResponse:
        return ledg_responses.FlashingResponse.from_json(json=json)


# LEDR factories
class LedrSetOnFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> ledr_responses.SetOnResponse:
        return ledr_responses.SetOnResponse.from_dpa(dpa)

    def create_from_json(self, json: dict) -> ledr_responses.SetOnResponse:
        return ledr_responses.SetOnResponse.from_json(json)


class LedrSetOffFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> ledr_responses.SetOffResponse:
        return ledr_responses.SetOffResponse.from_dpa(dpa)

    def create_from_json(self, json: dict) -> ledr_responses.SetOffResponse:
        return ledr_responses.SetOffResponse.from_json(json)


class LedrPulseFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> ledr_responses.PulseResponse:
        return ledr_responses.PulseResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> ledr_responses.PulseResponse:
        return ledr_responses.PulseResponse.from_json(json=json)


class LedrFlashingFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> ledr_responses.FlashingResponse:
        return ledr_responses.FlashingResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> ledr_responses.FlashingResponse:
        return ledr_responses.FlashingResponse.from_json(json=json)


# Node factories
class NodeReadFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> node_responses.ReadResponse:
        return node_responses.ReadResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> node_responses.ReadResponse:
        return node_responses.ReadResponse.from_json(json=json)


class NodeRemoveBondFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> node_responses.RemoveBondResponse:
        return node_responses.RemoveBondResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> node_responses.RemoveBondResponse:
        return node_responses.RemoveBondResponse.from_json(json=json)


class NodeBackupFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> node_responses.BackupResponse:
        return node_responses.BackupResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> node_responses.BackupResponse:
        return node_responses.BackupResponse.from_json(json=json)


class NodeRestoreFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> node_responses.RestoreResponse:
        return node_responses.RestoreResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> node_responses.RestoreResponse:
        return node_responses.RestoreResponse.from_json(json=json)


class NodeValidateBondsFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> node_responses.ValidateBondsResponse:
        return node_responses.ValidateBondsResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> node_responses.ValidateBondsResponse:
        return node_responses.ValidateBondsResponse.from_json(json=json)


# OS factories
class OSReadFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> os_responses.ReadResponse:
        return os_responses.ReadResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> os_responses.ReadResponse:
        return os_responses.ReadResponse.from_json(json=json)


class OSResetFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> os_responses.ResetResponse:
        return os_responses.ResetResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> os_responses.ResetResponse:
        return os_responses.ResetResponse.from_json(json=json)


class OSRestartFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> os_responses.RestartResponse:
        return os_responses.RestartResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> os_responses.RestartResponse:
        return os_responses.RestartResponse.from_json(json=json)


class OSReadTrConfFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> os_responses.ReadTrConfResponse:
        return os_responses.ReadTrConfResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> os_responses.ReadTrConfResponse:
        return os_responses.ReadTrConfResponse.from_json(json=json)


class OSWriteTrConfFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> os_responses.WriteTrConfResponse:
        return os_responses.WriteTrConfResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> os_responses.WriteTrConfResponse:
        return os_responses.WriteTrConfResponse.from_json(json=json)


class OSRfpgmFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> os_responses.RfpgmResponse:
        return os_responses.RfpgmResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> os_responses.RfpgmResponse:
        return os_responses.RfpgmResponse.from_json(json=json)


class OSSleepFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> os_responses.SleepResponse:
        return os_responses.SleepResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> os_responses.SleepResponse:
        return os_responses.SleepResponse.from_json(json=json)


class OSSetSecurityFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> os_responses.SetSecurityResponse:
        return os_responses.SetSecurityResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> os_responses.SetSecurityResponse:
        return os_responses.SetSecurityResponse.from_json(json=json)


# RAM factories
class RamReadFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> ram_responses.ReadResponse:
        return ram_responses.ReadResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> ram_responses.ReadResponse:
        return ram_responses.ReadResponse.from_json(json=json)


class RamWriteFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> ram_responses.WriteResponse:
        return ram_responses.WriteResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> ram_responses.WriteResponse:
        return ram_responses.WriteResponse.from_json(json=json)


class RamReadAnyFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> ram_responses.ReadAnyResponse:
        return ram_responses.ReadAnyResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> ram_responses.ReadAnyResponse:
        return ram_responses.ReadAnyResponse.from_json(json=json)


# Thermometer factories
class ThermometerReadFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> thermometer_responses.ReadResponse:
        return thermometer_responses.ReadResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> thermometer_responses.ReadResponse:
        return thermometer_responses.ReadResponse.from_json(json=json)


# UART factories
class UartOpenFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> uart_responses.OpenResponse:
        return uart_responses.OpenResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> uart_responses.OpenResponse:
        return uart_responses.OpenResponse.from_json(json=json)


class UartCloseFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> uart_responses.CloseResponse:
        return uart_responses.CloseResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> uart_responses.CloseResponse:
        return uart_responses.CloseResponse.from_json(json=json)


class UartWriteReadFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> uart_responses.WriteReadResponse:
        return uart_responses.WriteReadResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> uart_responses.WriteReadResponse:
        return uart_responses.WriteReadResponse.from_json(json=json)


class UartClearWriteReadFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> uart_responses.ClearWriteReadResponse:
        return uart_responses.ClearWriteReadResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> uart_responses.ClearWriteReadResponse:
        return uart_responses.ClearWriteReadResponse.from_json(json=json)


# BinaryOutput factories
class BinaryOutputEnumerateFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> bo_responses.EnumerateResponse:
        return bo_responses.EnumerateResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> bo_responses.EnumerateResponse:
        return bo_responses.EnumerateResponse.from_json(json=json)


class BinaryOutputSetFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> bo_responses.SetOutputResponse:
        return bo_responses.SetOutputResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> bo_responses.SetOutputResponse:
        return bo_responses.SetOutputResponse.from_json(json=json)


# Sensor factories
class SensorEnumerateFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> sensor_responses.EnumerateResponse:
        return sensor_responses.EnumerateResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> sensor_responses.EnumerateResponse:
        return sensor_responses.EnumerateResponse.from_json(json=json)


class SensorReadSensorsFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> sensor_responses.ReadSensorsResponse:
        return sensor_responses.ReadSensorsResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> sensor_responses.ReadSensorsResponse:
        return sensor_responses.ReadSensorsResponse.from_json(json=json)


class SensorReadSensorsWithTypesFactory(BaseFactory):

    def create_from_dpa(self, dpa: bytes) -> sensor_responses.ReadSensorsWithTypesResponse:
        return sensor_responses.ReadSensorsWithTypesResponse.from_dpa(dpa=dpa)

    def create_from_json(self, json: dict) -> sensor_responses.ReadSensorsWithTypesResponse:
        return sensor_responses.ReadSensorsWithTypesResponse.from_json(json=json)


class ResponseFactory:
    """Response factory class."""

    _dpa_factories: dict = {
        EmbedPeripherals.COORDINATOR: {
            CoordinatorResponseCommands.ADDR_INFO: CoordinatorAddrInfoFactory(),
            CoordinatorResponseCommands.AUTHORIZE_BOND: CoordinatorAuthorizeBondFactory(),
            CoordinatorResponseCommands.BACKUP: CoordinatorBackupFactory(),
            CoordinatorResponseCommands.BONDED_DEVICES: CoordinatorBondedDevicesFactory(),
            CoordinatorResponseCommands.BOND_NODE: CoordinatorBondNodeFactory(),
            CoordinatorResponseCommands.CLEAR_ALL_BONDS: CoordinatorClearAllBondsFactory(),
            CoordinatorResponseCommands.DISCOVERED_DEVICES: CoordinatorDiscoveredDevicesFactory(),
            CoordinatorResponseCommands.DISCOVERY: CoordinatorDiscoveryFactory(),
            CoordinatorResponseCommands.REMOVE_BOND: CoordinatorRemoveBondFactory(),
            CoordinatorResponseCommands.RESTORE: CoordinatorRestoreFactory(),
            CoordinatorResponseCommands.SET_DPA_PARAMS: CoordinatorSetDpaParamsFactory(),
            CoordinatorResponseCommands.SET_HOPS: CoordinatorSetHopsFactory(),
            CoordinatorResponseCommands.SET_MID: CoordinatorSetMIDFactory(),
            CoordinatorResponseCommands.SMART_CONNECT: CoordinatorSmartConnectFactory(),
        },
        EmbedPeripherals.EEEPROM: {
            EEEPROMResponseCommands.READ: EeepromReadFactory(),
            EEEPROMResponseCommands.WRITE: EeepromWriteFactory(),
        },
        EmbedPeripherals.EEPROM: {
            EEPROMResponseCommands.READ: EepromReadFactory(),
            EEPROMResponseCommands.WRITE: EepromWriteFactory(),
        },
        EmbedPeripherals.EXPLORATION: {
            ExplorationResponseCommands.PERIPHERALS_ENUMERATION_INFORMATION: ExplorationPeripheralEnumerationFactory(),
        },
        EmbedPeripherals.FRC: {
            FrcResponseCommands.SEND: FrcSendFactory(),
            FrcResponseCommands.EXTRA_RESULT: FrcExtraResultFactory(),
            FrcResponseCommands.SEND_SELECTIVE: FrcSendSelectiveFactory(),
            FrcResponseCommands.SET_PARAMS: FrcSetFrcParamsFactory(),
        },
        EmbedPeripherals.IO: {
            IOResponseCommands.DIRECTION: IoDirectionFactory(),
            IOResponseCommands.GET: IoGetFactory(),
            IOResponseCommands.SET: IoSetFactory(),
        },
        EmbedPeripherals.LEDG: {
            LEDResponseCommands.SET_ON: LedgSetOnFactory(),
            LEDResponseCommands.SET_OFF: LedgSetOffFactory(),
            LEDResponseCommands.PULSE: LedgPulseFactory(),
            LEDResponseCommands.FLASHING: LedgFlashingFactory(),
        },
        EmbedPeripherals.LEDR: {
            LEDResponseCommands.SET_ON: LedrSetOnFactory(),
            LEDResponseCommands.SET_OFF: LedrSetOffFactory(),
            LEDResponseCommands.PULSE: LedrPulseFactory(),
            LEDResponseCommands.FLASHING: LedrFlashingFactory(),
        },
        EmbedPeripherals.NODE: {
            NodeResponseCommands.READ: NodeReadFactory(),
            NodeResponseCommands.REMOVE_BOND: NodeRemoveBondFactory(),
            NodeResponseCommands.BACKUP: NodeBackupFactory(),
            NodeResponseCommands.RESTORE: NodeRestoreFactory(),
            NodeResponseCommands.VALIDATE_BONDS: NodeValidateBondsFactory(),
        },
        EmbedPeripherals.OS: {
            OSResponseCommands.READ: OSReadFactory(),
            OSResponseCommands.RESET: OSResetFactory(),
            OSResponseCommands.RESTART: OSRestartFactory(),
            OSResponseCommands.READ_CFG: OSReadTrConfFactory(),
            OSResponseCommands.WRITE_CFG: OSWriteTrConfFactory(),
            OSResponseCommands.RFPGM: OSRfpgmFactory(),
            OSResponseCommands.SLEEP: OSSleepFactory(),
            OSResponseCommands.SET_SECURITY: OSSetSecurityFactory(),
        },
        EmbedPeripherals.RAM: {
            RAMResponseCommands.READ: RamReadFactory(),
            RAMResponseCommands.WRITE: RamWriteFactory(),
            RAMResponseCommands.READ_ANY: RamReadAnyFactory(),
        },
        EmbedPeripherals.THERMOMETER: {
            ThermometerResponseCommands.READ: ThermometerReadFactory(),
        },
        EmbedPeripherals.UART: {
            UartResponseCommands.OPEN: UartOpenFactory(),
            UartResponseCommands.CLOSE: UartCloseFactory(),
            UartResponseCommands.WRITE_READ: UartWriteReadFactory(),
            UartResponseCommands.CLEAR_WRITE_READ: UartClearWriteReadFactory(),
        },
        Standards.BINARY_OUTPUT: {
            BinaryOutputResponseCommands.ENUMERATE_OUTPUTS: BinaryOutputEnumerateFactory(),
            BinaryOutputResponseCommands.SET_OUTPUT: BinaryOutputSetFactory(),
        },
        Standards.SENSOR: {
            SensorResponseCommands.ENUMERATE: SensorEnumerateFactory(),
            SensorResponseCommands.READ_SENSORS: SensorReadSensorsFactory(),
            SensorResponseCommands.READ_SENSORS_WITH_TYPES: SensorReadSensorsWithTypesFactory(),
        },
    }
    """DPA factories, dictionary (pnum) of dictionaries (pcmd)."""

    _json_factories: dict = {
        GenericMessages.RAW: GenericResponseFactory(),
        CoordinatorMessages.ADDR_INFO: CoordinatorAddrInfoFactory(),
        CoordinatorMessages.AUTHORIZE_BOND: CoordinatorAuthorizeBondFactory(),
        CoordinatorMessages.BACKUP: CoordinatorBackupFactory(),
        CoordinatorMessages.BONDED_DEVICES: CoordinatorBondedDevicesFactory(),
        CoordinatorMessages.BOND_NODE: CoordinatorBondNodeFactory(),
        CoordinatorMessages.CLEAR_ALL_BONDS: CoordinatorClearAllBondsFactory(),
        CoordinatorMessages.DISCOVERED_DEVICES: CoordinatorDiscoveredDevicesFactory(),
        CoordinatorMessages.DISCOVERY: CoordinatorDiscoveryFactory(),
        CoordinatorMessages.REMOVE_BOND: CoordinatorRemoveBondFactory(),
        CoordinatorMessages.RESTORE: CoordinatorRestoreFactory(),
        CoordinatorMessages.SET_DPA_PARAMS: CoordinatorSetDpaParamsFactory(),
        CoordinatorMessages.SET_HOPS: CoordinatorSetHopsFactory(),
        CoordinatorMessages.SET_MID: CoordinatorSetMIDFactory(),
        CoordinatorMessages.SMART_CONNECT: CoordinatorSmartConnectFactory(),
        EEEPROMMessages.READ: EeepromReadFactory(),
        EEEPROMMessages.WRITE: EeepromWriteFactory(),
        EEPROMMessages.READ: EepromReadFactory(),
        EEPROMMessages.WRITE: EepromWriteFactory(),
        ExplorationMessages.ENUMERATE: ExplorationPeripheralEnumerationFactory(),
        ExplorationMessages.PERIPHERAL_INFORMATION: ExplorationPeripheralInformationFactory(),
        ExplorationMessages.MORE_PERIPHERALS_INFORMATION: ExplorationMorePeripheralsInformationFactory(),
        FrcMessages.SEND: FrcSendFactory(),
        FrcMessages.EXTRA_RESULT: FrcExtraResultFactory(),
        FrcMessages.SEND_SELECTIVE: FrcSendSelectiveFactory(),
        FrcMessages.SET_PARAMS: FrcSetFrcParamsFactory(),
        IOMessages.DIRECTION: IoDirectionFactory(),
        IOMessages.GET: IoGetFactory(),
        IOMessages.SET: IoSetFactory(),
        LEDGMessages.SET_ON: LedgSetOnFactory(),
        LEDGMessages.SET_OFF: LedgSetOffFactory(),
        LEDGMessages.PULSE: LedgPulseFactory(),
        LEDGMessages.FLASHING: LedgFlashingFactory(),
        LEDRMessages.SET_ON: LedrSetOnFactory(),
        LEDRMessages.SET_OFF: LedrSetOffFactory(),
        LEDRMessages.PULSE: LedrPulseFactory(),
        LEDRMessages.FLASHING: LedrFlashingFactory(),
        NodeMessages.READ: NodeReadFactory(),
        NodeMessages.REMOVE_BOND: NodeRemoveBondFactory(),
        NodeMessages.BACKUP: NodeBackupFactory(),
        NodeMessages.RESTORE: NodeRestoreFactory(),
        NodeMessages.VALIDATE_BONDS: NodeValidateBondsFactory(),
        OSMessages.READ: OSReadFactory(),
        OSMessages.RESET: OSResetFactory(),
        OSMessages.RESTART: OSRestartFactory(),
        OSMessages.READ_CFG: OSReadTrConfFactory(),
        OSMessages.WRITE_CFG: OSWriteTrConfFactory(),
        OSMessages.RFPGM: OSRfpgmFactory(),
        OSMessages.SLEEP: OSSleepFactory(),
        OSMessages.SET_SECURITY: OSSetSecurityFactory(),
        RAMMessages.READ: RamReadFactory(),
        RAMMessages.WRITE: RamWriteFactory(),
        ThermometerMessages.READ: ThermometerReadFactory(),
        UartMessages.OPEN: UartOpenFactory(),
        UartMessages.CLOSE: UartCloseFactory(),
        UartMessages.WRITE_READ: UartWriteReadFactory(),
        UartMessages.CLEAR_WRITE_READ: UartClearWriteReadFactory(),
        BinaryOutputMessages.ENUMERATE: BinaryOutputEnumerateFactory(),
        BinaryOutputMessages.SET_OUTPUT: BinaryOutputSetFactory(),
        SensorMessages.ENUMERATE: SensorEnumerateFactory(),
        SensorMessages.READ_SENSORS_WITH_TYPES: SensorReadSensorsWithTypesFactory(),
    }
    """JSON API factories, dictionary (message types)."""

    @classmethod
    def register_dpa_response_factory(cls, pnum: Union[Peripheral, int], pcmd: Union[Command, int], factory: BaseFactory):
        """Register new DPA response factory.

        NOTE THAT THIS FEATURE IS A WORK-IN-PROGRESS.

        The new factory will be identified by a unique combination of peripheral number and peripheral command.

        Args:
            pnum (Union[Peripheral, int]): Peripheral number.
            pcmd (Union[Command, int]): Peripheral command.
            factory (BaseFactory): Response factory.

        Raises:
            ResponseFactoryRegisterError: If combination of pnum and pcmd already corresponds to an existing
                response factory.
        """
        if pnum in cls._dpa_factories:
            if pcmd in cls._dpa_factories[pnum]:
                raise ResponseFactoryRegisterError(f'A response factory is already registered under peripheral {pnum}'
                                                   f' and command {pcmd}.')
        cls._dpa_factories[pnum][pcmd] = factory

    @classmethod
    def unregister_dpa_response_factory(cls, pnum: Union[Peripheral, int], pcmd: Union[Command, int]):
        """Unregister DPA response factory.

        NOTE THAT THIS FEATURE IS A WORK-IN-PROGRESS.

        Response factory identified by a unique combination of peripheral number and peripheral command
            will be unregistered.

        Args:
            pnum (Union[Peripheral, int]): Peripheral number.
            pcmd (Union[Command, int]): Peripheral command.
        """
        if pnum in cls._dpa_factories and pcmd in cls._dpa_factories[pnum]:
            del cls._dpa_factories[pnum][pcmd]

    @classmethod
    def register_json_response_factory(cls, message_type: Union[MessageType, str], factory: BaseFactory):
        """Register new DPA response factory.

        NOTE THAT THIS FEATURE IS A WORK-IN-PROGRESS.

        The new factory will be identified by a unique message type.

        Args:
            message_type (Union[MessageType, str]): Message type.
            factory (BaseFactory): Response factory.

        Raises:
            ResponseFactoryRegisterError: If message_type already corresponds to an existing response factory.
        """
        if message_type in cls._json_factories:
            raise ResponseFactoryRegisterError(f'A response factory is already registered under message'
                                               f'type {message_type}.')
        cls._json_factories[message_type] = factory

    @classmethod
    def unregister_json_response_factory(cls, message_type: Union[MessageType, str]):
        """Unregister DPA response factory.

        NOTE THAT THIS FEATURE IS A WORK-IN-PROGRESS.

        Response factory identified by a unique message type will be unregistered.

        Args:
            message_type (Union[MessageType, str]): Message type.
        """
        if message_type in cls._json_factories:
            del cls._json_factories[message_type]

    @classmethod
    def get_factory_from_dpa(cls, pnum: Union[Peripheral, int], pcmd: Union[Command, int],
                             allow_generic: bool = False) -> BaseFactory:
        """Find response factory by combination of pnum and pcmd.

        Args:
            pnum (Union[Peripheral int]): Peripheral number.
            pcmd (Union[Command, int]): Peripheral command.
            allow_generic (bool): Use generic response factory if a matching factory for pnum and pcmd does not exist.
        Returns:
            :obj:`BaseFactory`: Response factory.
        Raises:
            UnsupportedPeripheralError: If pnum does not exist in response factories and allow_generic is False.
            UnsupportedPeripheralCommandError: If pcmd does not exist in response factories and allow_generic is False.
        """
        if pnum in cls._dpa_factories:
            if pcmd in cls._dpa_factories[pnum]:
                return cls._dpa_factories[pnum][pcmd]
            else:
                if allow_generic:
                    return GenericResponseFactory()
                else:
                    raise UnsupportedPeripheralCommandError(f'Unknown or unsupported peripheral command: {pcmd}')
        else:
            if allow_generic:
                return GenericResponseFactory()
            else:
                raise UnsupportedPeripheralError(f'Unknown or unsupported peripheral: {pnum}')

    @classmethod
    def get_factory_from_mtype(cls, message_type: Union[MessageType, str], msgid: Optional[str] = None) -> BaseFactory:
        """Find response factory by message type.

        Args:
            message_type (Union[MessageType, str): Message type.
            msgid (str, optional): Message ID, used for error handling.
        Returns:
            :obj:`BaseFactory`: Response factory.
        Raises:
            UnsupportedMessageTypeError: If message type does not exist in response factories.
        """
        if message_type in cls._json_factories:
            return cls._json_factories[message_type]
        raise UnsupportedMessageTypeError(f'Unknown or unsupported message type: {message_type}', msgid)

    @classmethod
    def get_response_from_dpa(cls, dpa: bytes) -> IResponse:
        """Process DPA response and return corresponding response message object.

        Args:
            dpa (bytes): DPA response.
        Returns:
            :obj:`IResponse`: Response object.
        """
        IResponse.validate_dpa_response(dpa)
        pnum = dpa[ResponsePacketMembers.PNUM]
        pcmd = dpa[ResponsePacketMembers.PCMD]
        rcode = dpa[ResponsePacketMembers.RCODE]
        if rcode == ResponseCodes.CONFIRMATION and len(dpa) == CONFIRMATION_PACKET_LEN:
            factory = ConfirmationFactory()
        elif pcmd <= REQUEST_PCMD_MAX and rcode >= ResponseCodes.ASYNC_RESPONSE:
            factory = AsyncResponseFactory()
        elif EmbedPeripherals.has_value(pnum) and pnum != EmbedPeripherals.EXPLORATION and \
                pcmd == ExplorationResponseCommands.PERIPHERALS_ENUMERATION_INFORMATION:
            factory = ExplorationPeripheralInformationFactory()
        elif pnum == BYTE_MAX and EmbedPeripherals.has_value(pcmd):
            factory = ExplorationMorePeripheralsInformationFactory()
            pass
        else:
            peripheral = Common.pnum_from_dpa(pnum)
            command = Common.response_pcmd_from_dpa(peripheral, pcmd)
            factory = cls.get_factory_from_dpa(peripheral, command)
        return factory.create_from_dpa(dpa)

    @classmethod
    def get_response_from_json(cls, json: dict) -> IResponse:
        """Process JSON API response and return corresponding response message object.

        Args:
            json (dict): JSON API response.
        Response:
            :obj:`IResponse`: Response object.
        """
        msgid = Common.msgid_from_json(json)
        mtype = Common.mtype_str_from_json(json)
        if msgid == IResponse.ASYNC_MSGID and \
                GenericMessages.has_value(mtype) and GenericMessages(mtype) == GenericMessages.RAW:
            factory = AsyncResponseFactory()
        else:
            message = Common.string_to_mtype(mtype)
            factory = cls.get_factory_from_mtype(message, msgid)
        return factory.create_from_json(json)



from iqrfpy.enums.message_types import MessageType


def generate_json_response(response_data: dict):
    json = {
        'mType': response_data['mtype'].value if issubclass(type(response_data['mtype']), MessageType) else response_data['mtype'],
        'data': {
            'msgId': response_data['msgid'],
            'rsp': {
                'nAdr': response_data['nadr'],
                'hwpId': response_data['hwpid'],
                'rCode': response_data['rcode'],
                'dpaVal': response_data['dpa_value']
            },
            'insId': 'iqrfgd2-1',
            'status': response_data['rcode']
        }
    }
    if 'pnum' in response_data:
        json['data']['rsp']['pnum'] = response_data['pnum']
    if 'pcmd' in response_data:
        json['data']['rsp']['pcmd'] = response_data['pcmd']
    if response_data['rcode'] == 0 and 'result' in response_data:
        json['data']['rsp']['result'] = response_data['result']
    return json



"""Exploration More Peripherals Information response message."""

from typing import List, Optional, Union
from iqrfpy.iresponse import IResponseGetterMixin
from iqrfpy.enums.commands import ExplorationResponsePeripheralCommand
from iqrfpy.enums.message_types import ExplorationMessages
from iqrfpy.enums.peripherals import EmbedPeripherals, Peripheral
from iqrfpy.exceptions import DpaResponsePacketLengthError
from iqrfpy.utils.common import Common
from iqrfpy.utils import dpa as dpa_constants
from iqrfpy.utils.dpa import ResponsePacketMembers, ResponseCodes
from iqrfpy.utils.validators import DpaValidator, JsonValidator
from iqrfpy.peripherals.exploration.responses.peripheral_information_data import PeripheralInformationData

__all__ = [
    'MorePeripheralsInformationResponse',
]


class MorePeripheralsInformationResponse(IResponseGetterMixin):
    """Exploration More Peripherals Information response class."""

    __slots__ = ('_peripheral_data',)

    def __init__(self, nadr: int, per: Union[Peripheral, int], hwpid: int = dpa_constants.HWPID_MAX,
                 rcode: int = 0, dpa_value: int = 0, msgid: Optional[str] = None, pdata: Optional[List[int]] = None,
                 result: Optional[dict] = None):
        """Read response constructor.

        Args:
            nadr (int): Device address.
            per (Union[Peripheral, int]): Requested peripheral.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535, this value ignores HWPID check.
            rcode (int, optional): Response code. Defaults to 128.
            dpa_value (int, optional): DPA value. Defaults to 0.
            pdata (List[int], optional): DPA response data. Defaults to None.
            msgid (str, optional): Message ID. Defaults to None.
            result (dict, optional): JSON response data. Defaults to None.
        """
        super().__init__(
            nadr=nadr,
            pnum=EmbedPeripherals.EXPLORATION,
            pcmd=per.value if isinstance(per, Peripheral) else per,
            m_type=ExplorationMessages.MORE_PERIPHERALS_INFORMATION,
            hwpid=hwpid,
            rcode=rcode,
            dpa_value=dpa_value,
            msgid=msgid,
            pdata=pdata,
            result=result
        )
        self._peripheral_data = None
        if rcode == ResponseCodes.OK:
            self._peripheral_data = [PeripheralInformationData(result=x) for x in result['peripherals']]

    @property
    def peripheral_data(self) -> Optional[List[PeripheralInformationData]]:
        """:obj:`list` of :obj:`PeripheralInformationData` or :obj:`None`: Peripheral information data.

        Getter only.
        """
        return self._peripheral_data

    @classmethod
    def from_dpa(cls, dpa: bytes) -> 'MorePeripheralsInformationResponse':
        """DPA response factory method.

        Parses DPA data and constructs MorePeripheralsInformationResponse object.

        Args:
            dpa (bytes): DPA response bytes.
        Returns:
            MorePeripheralsInformationResponse: Response message object.
        """
        DpaValidator.base_response_length(dpa=dpa)
        nadr = dpa[ResponsePacketMembers.NADR]
        hwpid = Common.hwpid_from_dpa(dpa[ResponsePacketMembers.HWPID_HI], dpa[ResponsePacketMembers.HWPID_LO])
        rcode = dpa[ResponsePacketMembers.RCODE]
        dpa_value = dpa[ResponsePacketMembers.DPA_VALUE]
        pdata = None
        result = None
        per = ExplorationResponsePeripheralCommand(dpa[ResponsePacketMembers.PCMD])
        if rcode == ResponseCodes.OK:
            pdata = Common.pdata_from_dpa(dpa=dpa)
            if len(pdata) % 4 != 0:
                raise DpaResponsePacketLengthError('Invalid DPA response length, PDATA should be in multiples of 4B.')
            result = {'peripherals': []}
            for i in range(0, len(pdata), 4):
                result['peripherals'].append({
                    'perTe': pdata[i],
                    'perT': pdata[i + 1],
                    'par1': pdata[i + 2],
                    'par2': pdata[i + 3]
                })
        return cls(nadr=nadr, per=per, hwpid=hwpid, rcode=rcode, dpa_value=dpa_value, pdata=pdata, result=result)

    @classmethod
    def from_json(cls, json: dict) -> 'MorePeripheralsInformationResponse':
        """JSON response factory method.

        Parses JSON API response and constructs MorePeripheralsInformationResponse object.

        Args:
            json (dict): JSON API response.
        Returns:
            MorePeripheralsInformationResponse: Response message object.
        """
        JsonValidator.response_received(json=json)
        nadr = Common.nadr_from_json(json=json)
        per = Common.pcmd_from_json(json=json)
        msgid = Common.msgid_from_json(json=json)
        hwpid = Common.hwpid_from_json(json=json)
        dpa_value = Common.dpa_value_from_json(json=json)
        rcode = Common.rcode_from_json(json=json)
        pdata = Common.pdata_from_json(json=json)
        result = Common.result_from_json(json=json) if rcode == ResponseCodes.OK else None
        if rcode < 0:
            per += 0x80
        return cls(nadr=nadr, per=per, msgid=msgid, hwpid=hwpid, rcode=rcode, dpa_value=dpa_value, pdata=pdata,
                   result=result)



"""Messages module.

This module serves as a single import for all messages.
"""

from iqrfpy.irequest import IRequest
from iqrfpy.iresponse import IResponse
from iqrfpy.async_response import AsyncResponse
from iqrfpy.confirmation import Confirmation

from iqrfpy.peripherals.generic.requests import GenericRequest
from iqrfpy.peripherals.generic.responses import GenericResponse

from iqrfpy.peripherals.coordinator.requests import AddrInfoRequest as CoordinatorAddrInfoReq, \
    AuthorizeBondRequest as CoordinatorAuthorizeBondReq, AuthorizeBondParams as CoordinatorAuthorizeBondParams, \
    BackupRequest as CoordinatorBackupReq, BondNodeRequest as CoordinatorBondNodeReq, \
    BondedDevicesRequest as CoordinatorBondedDevicesReq, ClearAllBondsRequest as CoordinatorClearAllBondsReq, \
    DiscoveredDevicesRequest as CoordinatorDiscoveredDevicesReq, DiscoveryRequest as CoordinatorDiscoveryReq, \
    RemoveBondRequest as CoordinatorRemoveBondReq, RestoreRequest as CoordinatorRestoreReq, \
    SetDpaParamsRequest as CoordinatorSetDpaParamsReq, DpaParam as CoordinatorDpaParam, \
    SetHopsRequest as CoordinatorSetHopsReq, SetMidRequest as CoordinatorSetMidReq, \
    SmartConnectRequest as CoordinatorSmartConnectReq
from iqrfpy.peripherals.coordinator.responses import AddrInfoResponse as CoordinatorAddrInfoRsp, \
    AuthorizeBondResponse as CoordinatorAuthorizeBondRsp, BackupResponse as CoordinatorBackupRsp, \
    BondNodeResponse as CoordinatorBondNodeRsp, BondedDevicesResponse as CoordinatorBondedDevicesRsp, \
    ClearAllBondsResponse as CoordinatorClearAllBondsRsp, DiscoveredDevicesResponse as CoordinatorDiscoveredDevicesRsp, \
    DiscoveryResponse as CoordinatorDiscoveryRsp, RemoveBondResponse as CoordinatorRemoveBondRsp, \
    RestoreResponse as CoordinatorRestoreRsp, SetDpaParamsResponse as CoordinatorSetDpaParamsRsp, \
    SetHopsResponse as CoordinatorSetHopsRsp, SetMidResponse as CoordinatorSetMidRsp, \
    SmartConnectResponse as CoordinatorSmartConnectRsp

from iqrfpy.peripherals.eeeprom.requests import ReadRequest as EeepromReadReq, WriteRequest as EeepromWriteReq
from iqrfpy.peripherals.eeeprom.responses import ReadResponse as EeepromReadRsp, WriteResponse as EeepromWriteRsp

from iqrfpy.peripherals.eeprom.requests import ReadRequest as EepromReadReq, WriteRequest as EepromWriteReq
from iqrfpy.peripherals.eeprom.responses import ReadResponse as EepromReadRsp, WriteResponse as EepromWriteRsp

from iqrfpy.peripherals.exploration.requests import PeripheralEnumerationRequest as \
    ExplorationPeripheralEnumerationReq, PeripheralInformationRequest as ExplorationPeripheralInformationReq, \
    MorePeripheralsInformationRequest as ExplorationMorePeripheralsInformationReq
from iqrfpy.peripherals.exploration.responses import PeripheralEnumerationResponse as \
    ExplorationPeripheralEnumerationRsp, PeripheralInformationResponse as ExplorationPeripheralInformationRsp, \
    MorePeripheralsInformationResponse as ExplorationMorePeripheralsInformationRsp
from iqrfpy.peripherals.exploration.responses.peripheral_information_data import PeripheralInformationData \
    as ExplorePeripheralInformationData
from iqrfpy.peripherals.exploration.responses.peripheral_enumeration_data import PeripheralEnumerationData \
    as ExplorePeripheralEnumerationData

from iqrfpy.peripherals.frc.requests import SendRequest as FrcSendReq, ExtraResultRequest as FrcExtraResultReq, \
    SendSelectiveRequest as FrcSendSelectiveReq, SetFrcParamsRequest as FrcSetFrcParamsReq, FrcParams
from iqrfpy.peripherals.frc.responses import SendResponse as FrcSendRsp, ExtraResultResponse as FrcExtraResultRsp, \
    SendSelectiveResponse as FrcSendSelectiveRsp, SetFrcParamsResponse as FrcSetFrcParamsRsp

from iqrfpy.peripherals.io.requests import DirectionRequest as IoDirectionReq, GetRequest as IoGetReq, \
    SetRequest as IoSetReq, IoTriplet
from iqrfpy.peripherals.io.responses import DirectionResponse as IoDirectionRsp, GetResponse as IoGetRsp, \
    SetResponse as IoSetRsp

from iqrfpy.peripherals.ledg.requests import SetOnRequest as LedgSetOnReq, SetOffRequest as LedgSetOffReq, \
    PulseRequest as LedgPulseReq, FlashingRequest as LedgFlashingReq
from iqrfpy.peripherals.ledg.responses import SetOnResponse as LedgSetOnRsp, SetOffResponse as LedgSetOffRsp, \
    PulseResponse as LedgPulseRsp, FlashingResponse as LedgFlashingRsp

from iqrfpy.peripherals.ledr.requests import SetOnRequest as LedrSetOnReq, SetOffRequest as LedrSetOffReq, \
    PulseRequest as LedrPulseReq, FlashingRequest as LedrFlashingReq
from iqrfpy.peripherals.ledr.responses import SetOnResponse as LedrSetOnRsp, SetOffResponse as LedrSetOffRsp, \
    PulseResponse as LedrPulseRsp, FlashingResponse as LedrFlashingRsp

from iqrfpy.peripherals.node.requests import ReadRequest as NodeReadReq, RemoveBondRequest as NodeRemoveBondReq, \
    BackupRequest as NodeBackupReq, RestoreRequest as NodeRestoreReq, ValidateBondsRequest as NodeValidateBondsReq, \
    NodeValidateBondsParams
from iqrfpy.peripherals.node.responses import ReadResponse as NodeReadRsp, NodeReadData, \
    RemoveBondResponse as NodeRemoveBondRsp, BackupResponse as NodeBackupRsp, RestoreResponse as NodeRestoreRsp, \
    ValidateBondsResponse as NodeValidateBondsRsp

from iqrfpy.peripherals.os.requests import ReadRequest as OsReadReq, ResetRequest as OsResetReq, \
    RestartRequest as OsRestartReq, ReadTrConfRequest as OsReadTrConfReq, WriteTrConfRequest as OsWriteTrConfReq, \
    RfpgmRequest as OsRfpgmReq, SleepRequest as OsSleepReq, OsSleepParams, SetSecurityRequest as OsSetSecurityReq, \
    OsSecurityType
from iqrfpy.peripherals.os.responses import ReadResponse as OsReadRsp, ResetResponse as OsResetRsp, \
    RestartResponse as OsRestartRsp, ReadTrConfResponse as OsReadTrConfRsp, WriteTrConfResponse as OsWriteTrConfRsp, \
    RfpgmResponse as OsRfpgmRsp, SleepResponse as OsSleepRsp, SetSecurityResponse as OsSetSecurityRsp
from iqrfpy.peripherals.os.os_tr_conf_data import OsTrConfData
from iqrfpy.peripherals.os.responses.os_read_data import OsReadData

from iqrfpy.peripherals.ram.requests import ReadRequest as RamReadReq, WriteRequest as RamWriteReq, \
    ReadAnyRequest as RamReadAnyReq
from iqrfpy.peripherals.ram.responses import ReadResponse as RamReadRsp, WriteResponse as RamWriteRsp, \
    ReadAnyResponse as RamReadAnyRsp

from iqrfpy.peripherals.thermometer.requests.read import ReadRequest as ThermometerReadReq
from iqrfpy.peripherals.thermometer.responses.read import ReadResponse as ThermometerReadRsp

from iqrfpy.peripherals.uart.requests import OpenRequest as UartOpenReq, CloseRequest as UartCloseReq, \
    WriteReadRequest as UartWriteReadReq, ClearWriteReadRequest as UartClearWriteReadReq
from iqrfpy.peripherals.uart.responses import OpenResponse as UartOpenRsp, CloseResponse as UartCloseRsp, \
    WriteReadResponse as UartWriteReadRsp, ClearWriteReadResponse as UartClearWriteReadRsp

from iqrfpy.peripherals.binaryoutput.requests import EnumerateRequest as BinaryOutputEnumerateReq, SetOutputRequest \
    as BinaryOutputSetOutputReq, BinaryOutputState
from iqrfpy.peripherals.binaryoutput.responses import EnumerateResponse as BinaryOutputEnumerateRsp, SetOutputResponse \
    as BinaryOutputSetOutputRsp

from iqrfpy.peripherals.sensor.requests import EnumerateRequest as SensorEnumerateReq, \
    ReadSensorsRequest as SensorReadSensorsReq, ReadSensorsWithTypesRequest as SensorReadWithTypesReq
from iqrfpy.peripherals.sensor.responses import EnumerateResponse as SensorEnumerateRsp, \
    ReadSensorsResponse as SensorReadSensorsRsp, ReadSensorsWithTypesResponse as SensorReadWithTypesRsp
from iqrfpy.peripherals.sensor.requests.sensor_written_data import SensorWrittenData
from iqrfpy.utils.sensor_parser import SensorData


__all__ = (
    'IRequest',
    'IResponse',
    'AsyncResponse',
    'Confirmation',
    'GenericRequest',
    'GenericResponse',
    'CoordinatorAddrInfoReq',
    'CoordinatorAddrInfoRsp',
    'CoordinatorAuthorizeBondParams',
    'CoordinatorAuthorizeBondReq',
    'CoordinatorAuthorizeBondRsp',
    'CoordinatorBackupReq',
    'CoordinatorBackupRsp',
    'CoordinatorBondNodeReq',
    'CoordinatorBondNodeRsp',
    'CoordinatorBondedDevicesReq',
    'CoordinatorBondedDevicesRsp',
    'CoordinatorClearAllBondsReq',
    'CoordinatorClearAllBondsRsp',
    'CoordinatorDiscoveredDevicesReq',
    'CoordinatorDiscoveredDevicesRsp',
    'CoordinatorDiscoveryReq',
    'CoordinatorDiscoveryRsp',
    'CoordinatorRemoveBondReq',
    'CoordinatorRemoveBondRsp',
    'CoordinatorRestoreReq',
    'CoordinatorRestoreRsp',
    'CoordinatorSetDpaParamsReq',
    'CoordinatorDpaParam',
    'CoordinatorSetDpaParamsRsp',
    'CoordinatorSetHopsReq',
    'CoordinatorSetHopsRsp',
    'CoordinatorSetMidReq',
    'CoordinatorSetMidRsp',
    'CoordinatorSmartConnectReq',
    'CoordinatorSmartConnectRsp',
    'EeepromReadReq',
    'EeepromReadRsp',
    'EeepromWriteReq',
    'EeepromWriteRsp',
    'EepromReadReq',
    'EepromReadRsp',
    'EepromWriteReq',
    'EepromWriteRsp',
    'ExplorePeripheralEnumerationData',
    'ExplorationPeripheralEnumerationReq',
    'ExplorationPeripheralEnumerationRsp',
    'ExplorePeripheralInformationData',
    'ExplorationPeripheralInformationReq',
    'ExplorationPeripheralInformationRsp',
    'ExplorationMorePeripheralsInformationReq',
    'ExplorationMorePeripheralsInformationRsp',
    'FrcSendReq',
    'FrcSendRsp',
    'FrcExtraResultReq',
    'FrcExtraResultRsp',
    'FrcSendSelectiveReq',
    'FrcSendSelectiveRsp',
    'FrcParams',
    'FrcSetFrcParamsReq',
    'FrcSetFrcParamsRsp',
    'IoTriplet',
    'IoDirectionReq',
    'IoDirectionRsp',
    'IoGetReq',
    'IoGetRsp',
    'IoSetReq',
    'IoSetRsp',
    'LedgSetOnReq',
    'LedgSetOnRsp',
    'LedgSetOffReq',
    'LedgSetOffRsp',
    'LedgPulseReq',
    'LedgPulseRsp',
    'LedgFlashingReq',
    'LedgFlashingRsp',
    'LedrSetOnReq',
    'LedrSetOnRsp',
    'LedrSetOffReq',
    'LedrSetOffRsp',
    'LedrPulseReq',
    'LedrPulseRsp',
    'LedrFlashingReq',
    'LedrFlashingRsp',
    'NodeReadReq',
    'NodeReadData',
    'NodeReadRsp',
    'NodeRemoveBondReq',
    'NodeRemoveBondRsp',
    'NodeBackupReq',
    'NodeBackupRsp',
    'NodeRestoreReq',
    'NodeRestoreRsp',
    'NodeValidateBondsParams',
    'NodeValidateBondsReq',
    'NodeValidateBondsRsp',
    'OsReadReq',
    'OsReadData',
    'OsReadRsp',
    'OsResetReq',
    'OsResetRsp',
    'OsRestartReq',
    'OsRestartRsp',
    'OsTrConfData',
    'OsReadTrConfReq',
    'OsReadTrConfRsp',
    'OsWriteTrConfReq',
    'OsWriteTrConfRsp',
    'OsRfpgmReq',
    'OsRfpgmRsp',
    'OsSleepReq',
    'OsSleepParams',
    'OsSleepRsp',
    'OsSecurityType',
    'OsSetSecurityReq',
    'OsSetSecurityRsp',
    'RamReadReq',
    'RamReadRsp',
    'RamWriteReq',
    'RamWriteRsp',
    'RamReadAnyReq',
    'RamReadAnyRsp',
    'ThermometerReadReq',
    'ThermometerReadRsp',
    'UartOpenReq',
    'UartOpenRsp',
    'UartCloseReq',
    'UartCloseRsp',
    'UartWriteReadReq',
    'UartWriteReadRsp',
    'UartClearWriteReadReq',
    'UartClearWriteReadRsp',
    'BinaryOutputEnumerateReq',
    'BinaryOutputEnumerateRsp',
    'BinaryOutputState',
    'BinaryOutputSetOutputReq',
    'BinaryOutputSetOutputRsp',
    'SensorData',
    'SensorEnumerateReq',
    'SensorEnumerateRsp',
    'SensorWrittenData',
    'SensorReadSensorsReq',
    'SensorReadSensorsRsp',
    'SensorReadWithTypesReq',
    'SensorReadWithTypesRsp',
)



"""Coordinator Discovery request message."""

from typing import Optional, Union
from iqrfpy.enums.commands import CoordinatorRequestCommands
from iqrfpy.enums.message_types import CoordinatorMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import RequestParameterInvalidValueError
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.irequest import IRequest

__all__ = ['DiscoveryRequest']


class DiscoveryRequest(IRequest):
    """Coordinator Discovery request class."""

    __slots__ = '_tx_power', '_max_addr'

    def __init__(self, tx_power: int, max_addr: int, hwpid: int = dpa_constants.HWPID_MAX,
                 dpa_rsp_time: Optional[float] = None, dev_process_time: Optional[float] = None,
                 msgid: Optional[str] = None):
        """Discovery request constructor

        Args:
            tx_power (int): TX power used for discovery.
            max_addr (int): Maximum node address to be included in the discovery process, 0 means all nodes.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate(tx_power, max_addr)
        super().__init__(
            nadr=dpa_constants.COORDINATOR_NADR,
            pnum=EmbedPeripherals.COORDINATOR,
            pcmd=CoordinatorRequestCommands.DISCOVERY,
            m_type=CoordinatorMessages.DISCOVERY,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._tx_power = tx_power
        self._max_addr = max_addr

    def _validate(self, tx_power: int, max_addr: int) -> None:
        """Validate request parameters.

        Addr:
            tx_power (int): TX power used for discovery.
            max_addr (int): Maximum node address to be included in the discovery process, 0 means all nodes.
        """
        self._validate_tx_power(tx_power)
        self._validate_max_addr(max_addr)

    @staticmethod
    def _validate_tx_power(tx_power: int):
        """Validate TX power parameter.

        Args:
            tx_power (int): TX power used for discovery.
        Raises:
            RequestParameterInvalidValueError: If tx_power is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= tx_power <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('TX power value should be between 0 and 255.')

    @property
    def tx_power(self) -> int:
        """:obj:`int`: TX power used for discovery.

        Getter and setter.
        """
        return self._tx_power

    @tx_power.setter
    def tx_power(self, value: int):
        self._validate_tx_power(value)
        self._tx_power = value

    @staticmethod
    def _validate_max_addr(max_addr: int):
        """Validate maximum node address parameter.

        Args:
            max_addr (int): Maximum node address to be included in the discovery process, 0 means all nodes.
        Raises:
            RequestParameterInvalidValueError: If max_addr is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= max_addr <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('Max address value should be between 0 and 255.')

    @property
    def max_addr(self) -> int:
        """:obj:`int`: Maximum node address to be included in the discovery process.

        Getter and setter.
        """
        return self._max_addr

    @max_addr.setter
    def max_addr(self, value: int) -> None:
        self._validate_max_addr(value)
        self._max_addr = value

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        self._pdata = [self._tx_power, self._max_addr]
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        self._params = {'txPower': self._tx_power, 'maxAddr': self._max_addr}
        return super().to_json()



"""DPA utility module.

This module contains DPA constants and enum classes.
"""

from .enums import IntEnumMember

__all__ = (
    'RequestPacketMembers',
    'ResponsePacketMembers',
    'COORDINATOR_NADR',
    'NADR_MIN',
    'NADR_MAX',
    'NODE_NADR_MIN',
    'NODE_NADR_MAX',
    'IQUIP_NADR',
    'PNUM_MAX',
    'REQUEST_PCMD_MIN',
    'REQUEST_PCMD_MAX',
    'RESPONSE_PCMD_MIN',
    'RESPONSE_PCMD_MAX',
    'HWPID_MIN',
    'HWPID_MAX',
    'CONFIRMATION_PACKET_LEN',
    'RESPONSE_GENERAL_LEN',
    'THERMOMETER_SENSOR_ERROR',
    'THERMOMETER_RESOLUTION',
    'MID_MIN',
    'MID_MAX',
    'LOCAL_DEVICE_ADDR',
    'IQMESH_TEMP_ADDR',
    'BROADCAST_ADDR',
    'BYTE_MIN',
    'BYTE_MAX',
    'WORD_MIN',
    'WORD_MAX',
    'PDATA_MAX_LEN',
    'EEEPROM_WRITE_MAX_DATA_LEN',
    'EEPROM_WRITE_MAX_DATA_LEN',
    'RAM_WRITE_MAX_DATA_LEN',
    'BACKUP_DATA_BLOCK_MAX_LEN',
    'BINOUT_INDEX_MIN',
    'BINOUT_INDEX_MAX',
    'SENSOR_INDEX_MIN',
    'SENSOR_INDEX_MAX',
    'BaudRates',
    'ResponseCodes',
    'PeripheralTypes',
    'ExtendedPeripheralCharacteristics',
    'FrcCommands',
    'UserFrcCommands',
    'FrcResponseTimes',
    'IoConstants'
)


class RequestPacketMembers(IntEnumMember):
    """Request packet member indices."""

    NADR = 0
    PNUM = 2
    PCMD = 3
    HWPID_LO = 4
    HWPID_HI = 5


class ResponsePacketMembers(IntEnumMember):
    """Response packet member indices."""

    NADR = 0
    PNUM = 2
    PCMD = 3
    HWPID_LO = 4
    HWPID_HI = 5
    RCODE = 6
    DPA_VALUE = 7


# general constants
COORDINATOR_NADR = NADR_MIN = 0
NODE_NADR_MIN = 0x01
NADR_MAX = NODE_NADR_MAX = 0xEF
IQUIP_NADR = 0xF0
PNUM_MAX = 0x7F
REQUEST_PCMD_MIN = 0
REQUEST_PCMD_MAX = 0x7F
RESPONSE_PCMD_MIN = 0x80
RESPONSE_PCMD_MAX = 0xFF
HWPID_MIN = 0
HWPID_MAX = 0xFFFF

# confirmation constants
CONFIRMATION_PACKET_LEN = 11

# response constants
RESPONSE_GENERAL_LEN = 8

# thermometer constants
THERMOMETER_SENSOR_ERROR = 0x80
THERMOMETER_RESOLUTION = 0.0625

# mid constants
MID_MIN = 0
MID_MAX = 0xFFFFFFFF

# other constants
IBK_LEN = 16
LOCAL_DEVICE_ADDR = 0xFC
IQMESH_TEMP_ADDR = 0xFE
BROADCAST_ADDR = 0xFF
BYTE_MIN = 0
BYTE_MAX = 255
WORD_MIN = 0
WORD_MAX = 65535

PDATA_MAX_LEN = 56
EEPROM_WRITE_MAX_DATA_LEN = 55
EEEPROM_WRITE_MAX_DATA_LEN = 54
RAM_WRITE_MAX_DATA_LEN = 48
BACKUP_DATA_BLOCK_MAX_LEN = 49

SENSOR_INDEX_MIN = BINOUT_INDEX_MIN = 0
SENSOR_INDEX_MAX = BINOUT_INDEX_MAX = 31


class BaudRates(IntEnumMember):
    """UART baud rate constants."""

    B1200 = 0
    B2400 = 1
    B4800 = 2
    B9600 = 3
    B19200 = 4
    B38400 = 5
    B57600 = 6
    B115200 = 7
    B230400 = 8


# rcode constants
class ResponseCodes(IntEnumMember):
    """DPA response codes."""

    OK = 0
    ERROR_FAIL = 1
    ERROR_PCMD = 2
    ERROR_PNUM = 3
    ERROR_ADDR = 4
    ERROR_DATA_LEN = 5
    ERROR_DATA = 6
    ERROR_HWPID = 7
    ERROR_NADR = 8
    ERROR_IFACE_CUSTOM_HANDLER = 9
    ERROR_MISSING_CUSTOM_DPA_HANDLER = 10
    ERROR_USER_FROM = 0x20
    ERROR_USER_TO = 0x3F
    RESERVED_FLAG = 0x40
    ASYNC_RESPONSE = 0x80
    CONFIRMATION = 0xFF

    def __str__(self):
        """Convert self to representation of error code."""
        return self.to_string(self.value)

    @classmethod
    def to_string(cls, value: int):
        """Convert value to string representation of error code.

        Args:
            value (int): Value to convert
        Returns:
            :obj:`str`: String representation of error code
        """
        if not BYTE_MIN <= value <= BYTE_MAX:
            return 'Invalid DPA response code'
        if cls.ERROR_USER_FROM <= value <= cls.ERROR_USER_TO:
            return 'User error code'
        flags = []
        if value & 0x40:
            flags.append('reserved')
            value -= 0x40
        if value & 0x80:
            flags.append('async')
            value -= 0x80
        if value not in cls._value2member_map_:
            return 'Unknown DPA response code'
        val = cls(value)
        str_val = None
        match val:
            case cls.OK:
                str_val = 'No error'
            case cls.ERROR_FAIL:
                str_val = 'General fail'
            case cls.ERROR_PCMD:
                str_val = 'Incorrect PCMD'
            case cls.ERROR_PNUM:
                str_val = 'Incorrect PNUM or PCMD'
            case cls.ERROR_ADDR:
                str_val = 'Incorrect Address'
            case cls.ERROR_DATA_LEN:
                str_val = 'Incorrect Data length'
            case cls.ERROR_DATA:
                str_val = 'Incorrect Data'
            case cls.ERROR_HWPID:
                str_val = 'Incorrect HW Profile ID used'
            case cls.ERROR_NADR:
                str_val = 'Incorrect NADR'
            case cls.ERROR_IFACE_CUSTOM_HANDLER:
                str_val = 'Data from interface consumed by Custom DPA Handler'
            case cls.ERROR_MISSING_CUSTOM_DPA_HANDLER:
                str_val = 'Custom DPA Handler is missing'
            case cls.CONFIRMATION:
                str_val = 'DPA confirmation'
        flag_str = '' if len(flags) == 0 else ''.join(f' [{flag}]' for flag in flags)
        return f'{str_val}{flag_str}'


class PeripheralTypes(IntEnumMember):
    """Peripheral type constants."""

    PERIPHERAL_TYPE_DUMMY = 0
    PERIPHERAL_TYPE_COORDINATOR = 1
    PERIPHERAL_TYPE_NODE = 2
    PERIPHERAL_TYPE_OS = 3
    PERIPHERAL_TYPE_EEPROM = 4
    PERIPHERAL_TYPE_BLOCK_EEPROM = 5
    PERIPHERAL_TYPE_RAM = 6
    PERIPHERAL_TYPE_LED = 7
    PERIPHERAL_TYPE_SPI = 8
    PERIPHERAL_TYPE_IO = 9
    PERIPHERAL_TYPE_UART = 10
    PERIPHERAL_TYPE_THERMOMETER = 11
    PERIPHERAL_TYPE_FRC = 14


class ExtendedPeripheralCharacteristics(IntEnumMember):
    """Extended peripheral characteristics constants."""

    PERIPHERAL_TYPE_EXTENDED_DEFAULT = 0
    PERIPHERAL_TYPE_EXTENDED_READ = 1
    PERIPHERAL_TYPE_EXTENDED_WRITE = 2
    PERIPHERAL_TYPE_EXTENDED_READ_WRITE = 3


class FrcCommands(IntEnumMember):
    """FRC command intervals."""

    FRC_2BIT_FROM = 0
    FRC_2BIT_TO = 0x7F
    FRC_1BYTE_FROM = 0x80
    FRC_1BYTE_TO = 0xDF
    FRC_2BYTE_FROM = 0xE0
    FRC_2BYTE_TO = 0xF7
    FRC_4BYTE_FROM = 0xF8
    FRC_4BYTE_TO = 0xFF


class UserFrcCommands(IntEnumMember):
    """User FRC command intervals."""

    USER_BIT_FROM = 0x40
    USER_BIT_TO = 0x7F
    USER_BYTE_FROM = 0xC0
    USER_BYTE_TO = 0xDF
    USER_2BYTE_FROM = 0xF0
    USER_2BYTE_TO = 0xFF
    USER_4BYTE_FROM = 0xFC
    USER_4BYTE_TO = 0xFF


class FrcResponseTimes(IntEnumMember):
    """FRC response time constants."""

    MS40 = 0
    MS360 = 16
    MS680 = 32
    MS1320 = 48
    MS2600 = 64
    MS5160 = 80
    MS10280 = 96
    MS20520 = 112


class IoConstants(IntEnumMember):
    """IO peripheral constants enum."""

    TRIS_A = PORT_A = 0x00
    TRIS_B = PORT_B = 0x01
    TRIS_C = PORT_C = 0x02
    TRIS_E = PORT_E = 0x04
    PULL_UP_A = 0x10
    PULL_UP_B = 0x11
    PULL_UP_C = 0x12
    PULL_UP_E = 0x14
    DELAY = 0xFF



"""Sensor standard request messages."""

from .enumerate import EnumerateRequest
from .read_sensors import ReadSensorsRequest
from .read_sensors_with_types import ReadSensorsWithTypesRequest
from .sensor_written_data import SensorWrittenData

__all__ = [
    'EnumerateRequest',
    'SensorWrittenData',
    'ReadSensorsRequest',
    'ReadSensorsWithTypesRequest',
]



import random
import unittest
from typing import List
from parameterized import parameterized
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.peripherals.frc.requests.send_selective import SendSelectiveRequest


class SendSelectiveRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x00\x00\x0d\x02\xff\xff\x80\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                   b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        self.json = {
            'mType': 'iqrfEmbedFrc_SendSelective',
            'data': {
                'msgId': 'sendSelectiveTest',
                'req': {
                    'nAdr': 0,
                    'hwpId': 65535,
                    'param': {
                        'frcCommand': 128,
                        'selectedNodes': [1, 2, 3],
                        'userData': [0, 0]
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        [
            'ping',
            0,
            [1, 2, 3],
            [0, 0],
            b'\x00\x00\x0d\x02\xff\xff\x00\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        ],
        [
            'temperature',
            128,
            [1, 2, 3],
            [0, 0],
            b'\x00\x00\x0d\x02\xff\xff\x80\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        ],
        [
            'acknowledged_broadcast',
            2,
            [1, 2, 3],
            [6, 6, 3, 255, 255, 0],
            b'\x00\x00\x0d\x02\xff\xff\x02\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x06\x06\x03\xff\xff\x00'
        ]
    ])
    def test_to_dpa(self, _, frc_command: int, selected_nodes: List[int], user_data: List[int], expected):
        request = SendSelectiveRequest(nadr=0, frc_command=frc_command, selected_nodes=selected_nodes,
                                       user_data=user_data)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        ['ping', 0, [1, 2, 3], [0, 0]],
        ['temperature', 128, [1, 2, 3], [0, 0]],
        ['acknowledged_broadcast', 2, [1, 2, 3], [6, 6, 3, 255, 255, 0]]
    ])
    def test_to_json(self, _, frc_command: int, selected_nodes: List[int], user_data: List[int]):
        request = SendSelectiveRequest(nadr=0, frc_command=frc_command, selected_nodes=selected_nodes,
                                       user_data=user_data, msgid='sendSelectiveTest')
        self.json['data']['req']['param'] = {
            'frcCommand': frc_command,
            'selectedNodes': selected_nodes,
            'userData': user_data
        }
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [
            'ping',
            0,
            b'\x00\x00\x0d\x02\xff\xff\x00\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        ],
        [
            'acknowledged_broadcast',
            2,
            b'\x00\x00\x0d\x02\xff\xff\x02\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        ]
    ])
    def test_set_frc_command(self, _, frc_command: int, dpa):
        request = SendSelectiveRequest(nadr=0, frc_command=128, selected_nodes=[1, 2, 3], user_data=[0, 0],
                                       msgid='sendSelectiveTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.frc_command = frc_command
        self.json['data']['req']['param']['frcCommand'] = frc_command
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_set_frc_command_invalid(self, frc_command: int):
        request = SendSelectiveRequest(nadr=0, frc_command=128, selected_nodes=[1, 2, 3], user_data=[0, 0],
                                       msgid='sendSelectiveTest')
        with self.assertRaises(RequestParameterInvalidValueError):
            request.frc_command = frc_command

    @parameterized.expand([
        [
            [31, 32, 50],
            b'\x00\x00\x0d\x02\xff\xff\x80\x00\x00\x00\x80\x01\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        ],
        [
            [1, 2, 17, 50, 78],
            b'\x00\x00\x0d\x02\xff\xff\x80\x06\x00\x02\x00\x00\x00\x04\x00\x00\x40\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        ]
    ])
    def test_set_selected_nodes(self, selected_nodes: List[int], dpa):
        request = SendSelectiveRequest(nadr=0, frc_command=128, selected_nodes=[1, 2, 3], user_data=[0, 0],
                                       msgid='sendSelectiveTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.selected_nodes = selected_nodes
        self.json['data']['req']['param']['selectedNodes'] = selected_nodes
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [[random.randint(0, 255) for _ in range(250)]],
        [[-1]],
        [[240]],
        [[255]]
    ])
    def test_set_selected_nodes_invalid(self, selected_nodes: List[int]):
        request = SendSelectiveRequest(nadr=0, frc_command=128, selected_nodes=[1, 2, 3], user_data=[0, 0],
                                       msgid='sendSelectiveTest')
        with self.assertRaises(RequestParameterInvalidValueError):
            request.selected_nodes = selected_nodes

    @parameterized.expand([
        [
            'ping',
            [0, 0],
            b'\x00\x00\x0d\x02\xff\xff\x80\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        ],
        [
            'acknowledged_broadcast',
            [6, 6, 3, 255, 255, 0],
            b'\x00\x00\x0d\x02\xff\xff\x80\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x06\x06\x03\xff\xff\x00'
        ]
    ])
    def test_set_user_data(self, _, user_data: List[int], dpa):
        request = SendSelectiveRequest(nadr=0, frc_command=128, selected_nodes=[1, 2, 3], user_data=[0, 0],
                                       msgid='sendSelectiveTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.user_data = user_data
        self.json['data']['req']['param']['userData'] = user_data
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [[random.randint(0, 255) for _ in range(60)]],
        [[-1] * 27],
        [[256] * 27]
    ])
    def test_set_user_data_invalid(self, user_data: List[int]):
        request = SendSelectiveRequest(nadr=0, frc_command=128, selected_nodes=[1, 2, 3], user_data=[0, 0],
                                       msgid='sendSelectiveTest')
        with self.assertRaises(RequestParameterInvalidValueError):
            request.user_data = user_data



import unittest
from parameterized import parameterized
from iqrfpy.peripherals.coordinator.requests.remove_bond import RemoveBondRequest


class RemoveBondRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x00\x00\x00\x05\xff\xff\x01'
        self.json = {
            'mType': 'iqrfEmbedCoordinator_RemoveBond',
            'data': {
                'msgId': 'removeBondTest',
                'req': {
                    'nAdr': 0,
                    'hwpId': 65535,
                    'param': {
                        'bondAddr': 1
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        ['bond_addr', 10, b'\x00\x00\x00\x05\xff\xff\x0a'],
        ['bond_addr', 170, b'\x00\x00\x00\x05\xff\xff\xaa'],
    ])
    def test_to_dpa(self, _, bond_addr, expected):
        request = RemoveBondRequest(bond_addr=bond_addr)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        ['bond_addr', 10],
        ['bond_addr', 170],
    ])
    def test_to_json(self, _, bond_addr):
        request = RemoveBondRequest(bond_addr=bond_addr, msgid='removeBondTest')
        self.json['data']['req']['param']['bondAddr'] = bond_addr
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [5, b'\x00\x00\x00\x05\xff\xff\x05'],
        [239, b'\x00\x00\x00\x05\xff\xff\xef']
    ])
    def test_set_index(self, bond_addr, dpa):
        request = RemoveBondRequest(bond_addr=1, msgid='removeBondTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.bond_addr = bond_addr
        self.json['data']['req']['param']['bondAddr'] = bond_addr
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_construct_invalid(self, bond_addr):
        with self.assertRaises(ValueError):
            RemoveBondRequest(bond_addr=bond_addr)



"""iqrfpy

.. include:: ../README.md

.. include:: ../changelog.md
"""
__version__ = '0.1.40'

from typeguard import install_import_hook
with install_import_hook('iqrfpy'):
    from iqrfpy import async_response, confirmation, enums, exceptions, irequest, iresponse, messages, peripherals, \
        transports, utils, response_factory



"""Message types module.

This module contains embed peripherals and standards message type enums.
These enums are extended with has_value() method for member identification.
"""

import enum

__all__ = [
    'MessageType',
    'GenericMessages',
    'ExplorationMessages',
    'CoordinatorMessages',
    'NodeMessages',
    'OSMessages',
    'EEPROMMessages',
    'EEEPROMMessages',
    'RAMMessages',
    'LEDRMessages',
    'LEDGMessages',
    'IOMessages',
    'ThermometerMessages',
    'UartMessages',
    'FrcMessages',
    'DALIMessages',
    'BinaryOutputMessages',
    'SensorMessages',
    'LightMessages'
]


class MessageType(enum.Enum):
    """JSON API message type base enum."""

    @classmethod
    def has_value(cls, value: str):
        """Check if enum class has member value.

        Args:
            value (str): value to check
        Returns:
            :obj`bool`: True if value is a member of enum, False otherwise
        """
        return value in cls._value2member_map_


class GenericMessages(MessageType):
    """Generic message types enum."""

    RAW = 'iqrfRaw'
    RAW_HDP = 'iqrfRawHdp'


class ExplorationMessages(MessageType):
    """Exploration message types enum."""

    ENUMERATE = 'iqrfEmbedExplore_Enumerate'
    PERIPHERAL_INFORMATION = 'iqrfEmbedExplore_PeripheralInformation'
    MORE_PERIPHERALS_INFORMATION = 'iqrfEmbedExplore_MorePeripheralsInformation'


class CoordinatorMessages(MessageType):
    """Coordinator message types enum."""

    ADDR_INFO = 'iqrfEmbedCoordinator_AddrInfo'
    DISCOVERED_DEVICES = 'iqrfEmbedCoordinator_DiscoveredDevices'
    BONDED_DEVICES = 'iqrfEmbedCoordinator_BondedDevices'
    CLEAR_ALL_BONDS = 'iqrfEmbedCoordinator_ClearAllBonds'
    BOND_NODE = 'iqrfEmbedCoordinator_BondNode'
    REMOVE_BOND = 'iqrfEmbedCoordinator_RemoveBond'
    DISCOVERY = 'iqrfEmbedCoordinator_Discovery'
    SET_DPA_PARAMS = 'iqrfEmbedCoordinator_SetDpaParams'
    SET_HOPS = 'iqrfEmbedCoordinator_SetHops'
    BACKUP = 'iqrfEmbedCoordinator_Backup'
    RESTORE = 'iqrfEmbedCoordinator_Restore'
    AUTHORIZE_BOND = 'iqrfEmbedCoordinator_AuthorizeBond'
    SMART_CONNECT = 'iqrfEmbedCoordinator_SmartConnect'
    SET_MID = 'iqrfEmbedCoordinator_SetMID'


class NodeMessages(MessageType):
    """Node message types enum."""

    READ = 'iqrfEmbedNode_Read'
    REMOVE_BOND = 'iqrfEmbedNode_RemoveBond'
    BACKUP = 'iqrfEmbedNode_Backup'
    RESTORE = 'iqrfEmbedNode_Restore'
    VALIDATE_BONDS = 'iqrfEmbedNode_ValidateBonds'


class OSMessages(MessageType):
    """OS message types enum."""

    READ = 'iqrfEmbedOs_Read'
    RESET = 'iqrfEmbedOs_Reset'
    READ_CFG = 'iqrfEmbedOs_ReadCfg'
    RFPGM = 'iqrfEmbedOs_Rfpgm'
    SLEEP = 'iqrfEmbedOs_Sleep'
    BATCH = 'iqrfEmbedOs_Batch'
    SET_SECURITY = 'iqrfEmbedOs_SetSecurity'
    INDICATE = 'iqrfEmbedOs_Indicate'
    RESTART = 'iqrfEmbedOs_Restart'
    WRITE_CFG_BYTE = 'iqrfEmbedOs_WriteCfgByte'
    LOAD_CODE = 'iqrfEmbedOs_LoadCode'
    SELECTIVE_BATCH = 'iqrfEmbedOs_SelectiveBatch'
    TEST_RF_SIGNAL = 'iqrfEmbedOs_TestRfSignal'
    FACTORY_SETTINGS = 'iqrfEmbedOs_FactorySettings'
    WRITE_CFG = 'iqrfEmbedOs_WriteCfg'


class RAMMessages(MessageType):
    """RAM message types enum."""

    READ = 'iqrfEmbedRam_Read'
    WRITE = 'iqrfEmbedRam_Write'


class EEPROMMessages(MessageType):
    """EEPROM message types enum."""

    READ = 'iqrfEmbedEeprom_Read'
    WRITE = 'iqrfEmbedEeprom_Write'


class EEEPROMMessages(MessageType):
    """EEEPROM message types enum."""

    READ = 'iqrfEmbedEeeprom_Read'
    WRITE = 'iqrfEmbedEeeprom_Write'


class LEDRMessages(MessageType):
    """LEDR message types enum."""

    SET_ON = 'iqrfEmbedLedr_SetOn'
    SET_OFF = 'iqrfEmbedLedr_SetOff'
    PULSE = 'iqrfEmbedLedr_Pulse'
    FLASHING = 'iqrfEmbedLedr_Flashing'


class LEDGMessages(MessageType):
    """LEDG message types enum."""

    SET_ON = 'iqrfEmbedLedg_SetOn'
    SET_OFF = 'iqrfEmbedLedg_SetOff'
    PULSE = 'iqrfEmbedLedg_Pulse'
    FLASHING = 'iqrfEmbedLedg_Flashing'


class IOMessages(MessageType):
    """IO message types enum."""

    DIRECTION = 'iqrfEmbedIo_Direction'
    SET = 'iqrfEmbedIo_Set'
    GET = 'iqrfEmbedIo_Get'


class ThermometerMessages(MessageType):
    """READ message types enum."""

    READ = 'iqrfEmbedThermometer_Read'


class UartMessages(MessageType):
    """Uart message types enum."""

    OPEN = 'iqrfEmbedUart_Open'
    CLOSE = 'iqrfEmbedUart_Close'
    WRITE_READ = 'iqrfEmbedUart_WriteRead'
    CLEAR_WRITE_READ = 'iqrfEmbedUart_ClearWriteRead'


class FrcMessages(MessageType):
    """FRC message types enum."""

    SEND = 'iqrfEmbedFrc_Send'
    EXTRA_RESULT = 'iqrfEmbedFrc_ExtraResult'
    SEND_SELECTIVE = 'iqrfEmbedFrc_SendSelective'
    SET_PARAMS = 'iqrfEmbedFrc_SetParams'


class DALIMessages(MessageType):
    """DALI message types enum."""

    SEND_COMMANDS = 'iqrfDali_SendCommands'
    SEND_COMMANDS_ASYNC = 'iqrfDali_SendCommandsAsync'
    FRC = 'iqrfDali_Frc'


class BinaryOutputMessages(MessageType):
    """BinaryOutput message types enum."""

    SET_OUTPUT = 'iqrfBinaryoutput_SetOutput'
    ENUMERATE = 'iqrfBinaryoutput_Enumerate'


class SensorMessages(MessageType):
    """Sensor message types enum."""

    READ_SENSORS_WITH_TYPES = 'iqrfSensor_ReadSensorsWithTypes'
    ENUMERATE = 'iqrfSensor_Enumerate'
    FRC = 'iqrfSensor_Frc'


class LightMessages(MessageType):
    """Light message types enum."""

    SET_POWER = 'iqrfLight_SetPower'
    INCREMENT_POWER = 'iqrfLight_IncrementPower'
    DECREMENT_POWER = 'iqrfLight_DecrementPower'
    ENUMERATE = 'iqrfLight_Enumerate'



import unittest
from parameterized import parameterized
from iqrfpy.enums.commands import NodeResponseCommands
from iqrfpy.enums.message_types import NodeMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.peripherals.node.responses.backup import BackupResponse
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': NodeMessages.BACKUP,
    'msgid': 'backupTest',
    'nadr': 1,
    'hwpid': 1026,
    'rcode': 0,
    'dpa_value': 81,
    'result': {
        'backupData': [118, 141, 15, 207, 51, 118, 253, 211, 77, 1, 57, 187, 185, 27, 251, 200, 192, 106, 183, 80, 136,
                       142, 134, 92, 157, 69, 114, 33, 102, 98, 56, 230, 194, 61, 161, 152, 107, 233, 102, 228, 247,
                       134, 72, 37, 246, 216, 175, 188, 6]
    },
    'dpa': b'\x01\x00\x01\x86\x02\x04\x00\x51\x76\x8d\x0f\xcf\x33\x76\xfd\xd3\x4d\x01\x39\xbb\xb9\x1b\xfb\xc8\xc0\x6a'
           b'\xb7\x50\x88\x8e\x86\x5c\x9d\x45\x72\x21\x66\x62\x38\xe6\xc2\x3d\xa1\x98\x6b\xe9\x66\xe4\xf7\x86\x48\x25'
           b'\xf6\xd8\xaf\xbc\x06'
}

data_error: dict = {
    'mtype': NodeMessages.BACKUP,
    'msgid': 'backupTest',
    'nadr': 1,
    'hwpid': 1028,
    'rcode': 1,
    'dpa_value': 35,
    'dpa': b'\x01\x00\x01\x86\x04\x04\x01\x23'
}


class BackupResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, BackupResponse.from_dpa(data_ok['dpa']), False],
        ['from_json', data_ok, BackupResponse.from_json(generate_json_response(data_ok)), True],
        ['from_dpa_error', data_error, BackupResponse.from_dpa(data_error['dpa']), False],
        ['from_json_error', data_error, BackupResponse.from_json(generate_json_response(data_error)), True]
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, EmbedPeripherals.NODE)
        with self.subTest():
            self.assertEqual(response.pcmd, NodeResponseCommands.BACKUP)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, NodeMessages.BACKUP)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    def test_from_dpa_invalid(self):
        with self.assertRaises(ValueError):
            BackupResponse.from_dpa(b'\x01\x00\x01\x81\x00\x00\x00\x40')

    @parameterized.expand([
        ['from_dpa', data_ok['result']['backupData'], BackupResponse.from_dpa(data_ok['dpa'])],
        ['from_json', data_ok['result']['backupData'], BackupResponse.from_json(generate_json_response(data_ok))]
    ])
    def test_get_backup_data(self, _, backup_data, response: BackupResponse):
        self.assertEqual(response.backup_data, backup_data)

    @parameterized.expand([
        ['from_dpa_error', BackupResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', BackupResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_backup_data_error(self, _, response: BackupResponse):
        self.assertIsNone(response.backup_data)



"""Binary Output standard request messages."""

from .binary_output_state import BinaryOutputState
from .enumerate import EnumerateRequest
from .set_output import SetOutputRequest

__all__ = [
    'EnumerateRequest',
    'BinaryOutputState',
    'SetOutputRequest',
]



import unittest
from parameterized import parameterized
from iqrfpy.enums.commands import OSResponseCommands
from iqrfpy.enums.message_types import OSMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import DpaResponsePacketLengthError
from iqrfpy.peripherals.os.responses.sleep import SleepResponse
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': OSMessages.SLEEP,
    'msgid': 'sleepTest',
    'nadr': 2,
    'hwpid': 0,
    'rcode': 0,
    'dpa_value': 64,
    'dpa': b'\x02\x00\x02\x84\x00\x00\x00\x40'
}

data_error: dict = {
    'mtype': OSMessages.RESET,
    'msgid': 'resetTest',
    'nadr': 0,
    'hwpid': 1026,
    'rcode': 8,
    'dpa_value': 35,
    'dpa': b'\x00\x00\x02\x84\x02\x04\x08\x23'
}


class SleepResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, SleepResponse.from_dpa(data_ok['dpa']), False],
        ['from_json', data_ok, SleepResponse.from_json(generate_json_response(data_ok)), True],
        ['from_dpa_error', data_error, SleepResponse.from_dpa(data_error['dpa']), False],
        ['from_json_error', data_error, SleepResponse.from_json(generate_json_response(data_error)), True]
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, EmbedPeripherals.OS)
        with self.subTest():
            self.assertEqual(response.pcmd, OSResponseCommands.SLEEP)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, OSMessages.SLEEP)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    def test_from_dpa_invalid(self):
        with self.assertRaises(DpaResponsePacketLengthError):
            SleepResponse.from_dpa(b'\x02\x00\x02\x84\x00\x00\x00\x22\x32')



import random
import unittest
from typing import List
from parameterized import parameterized
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.peripherals.uart.requests.write_read import WriteReadRequest


class WriteReadRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x03\x00\x0c\x02\xff\xff\x01\x01\x02\x03\x04\x05'
        self.json = {
            'mType': 'iqrfEmbedUart_WriteRead',
            'data': {
                'msgId': 'writeReadTest',
                'req': {
                    'nAdr': 3,
                    'hwpId': 65535,
                    'param': {
                        'readTimeout': 1,
                        'writtenData': [1, 2, 3, 4, 5]
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        [random.randint(0, 255), random.sample(range(0, 255), 10)]
    ])
    def test_to_dpa(self, read_timeout: int, data: List[int]):
        expected = self.dpa[0:6] + bytes([read_timeout]) + bytes(data)
        request = WriteReadRequest(nadr=3, read_timeout=read_timeout, data=data)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        [random.randint(0, 255), random.sample(range(0, 255), 10)],
    ])
    def test_to_json(self, read_timeout: int, data: List[int]):
        request = WriteReadRequest(nadr=3, read_timeout=read_timeout, data=data, msgid='writeReadTest')
        self.json['data']['req']['param'] = {
            'readTimeout': read_timeout,
            'writtenData': data,
        }
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        ['read_timeout', 10],
        ['read_timeout', 20],
        ['read_timeout', 7],
    ])
    def test_set_read_timeout(self, _, read_timeout: int):
        request = WriteReadRequest(nadr=3, read_timeout=1, data=[1, 2, 3, 4, 5], msgid='writeReadTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        dpa = self.dpa[0:6] + bytes([read_timeout]) + self.dpa[7:]
        self.json['data']['req']['param']['readTimeout'] = read_timeout
        request.read_timeout = read_timeout
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        ['written_data', random.sample(range(0, 255), 10)],
    ])
    def test_set_write_data(self, _, data: List[int]):
        request = WriteReadRequest(nadr=3, read_timeout=1, data=[1, 2, 3, 4, 5], msgid='writeReadTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        dpa = self.dpa[0:7] + bytes(data)
        self.json['data']['req']['param']['writtenData'] = data
        request.data = data
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        ['negative_read_timeout', -1, [1]],
        ['read_timeout_word', 256, [1]],
        ['read_timeout_invalid', 1000, [1]],
        ['data_invalid', 1, [-1]],
        ['data_too_large', 1, random.sample(range(0, 255), 70)],
    ])
    def test_construct_invalid(self, _, read_timeout: int, data: List[int]):
        with self.assertRaises(RequestParameterInvalidValueError):
            WriteReadRequest(nadr=3, read_timeout=read_timeout, data=data)



"""Transports module containing transport abstract class and implementations for various communication channels.

Classes
-------
ITransport
MqttTransport
"""

from . import itransport
from . import mqtt_transport



import random
import unittest
from parameterized import parameterized
from typing import List
from iqrfpy.enums.message_types import BinaryOutputMessages
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.peripherals.binaryoutput.requests.binary_output_state import BinaryOutputState
from iqrfpy.peripherals.binaryoutput.requests.set_output import SetOutputRequest


class SetOutputRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x03\x00\x4b\x00\xff\xff\x00\x00\x00\x00'
        self.json = {
            'mType': BinaryOutputMessages.SET_OUTPUT.value,
            'data': {
                'msgId': 'setOutputTest',
                'req': {
                    'nAdr': 3,
                    'hwpId': 65535,
                    'param': {
                        'binOuts': [],
                    },
                },
                'returnVerbose': True,
            },
        }

    @parameterized.expand([
        [
            'single_pair',
            [
                BinaryOutputState(index=0, state=10)
            ],
            b'\x03\x00\x4b\x00\xff\xff\x01\x00\x00\x00\x0a'
        ],
        [
            'three_pairs',
            [
                BinaryOutputState(index=10, state=30),
                BinaryOutputState(index=5, state=10),
                BinaryOutputState(index=1, state=255)
            ],
            b'\x03\x00\x4b\x00\xff\xff\x22\x04\x00\x00\xff\x0a\x1e'
        ],
    ])
    def test_to_dpa(self, _, binouts: List[BinaryOutputState], expected: bytes):
        request = SetOutputRequest(nadr=3, binouts=binouts)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        [
            'single_pair',
            [
                BinaryOutputState(index=0, state=10)
            ],
        ],
        [
            'three_pairs',
            [
                BinaryOutputState(index=10, state=30),
                BinaryOutputState(index=5, state=10),
                BinaryOutputState(index=1, state=255)
            ],
        ],
    ])
    def test_to_json(self, _, binouts: List[BinaryOutputState]):
        request = SetOutputRequest(nadr=3, msgid='setOutputTest')
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.binouts = binouts
        self.json['data']['req']['param']['binOuts'] = [bo.to_json() for bo in binouts]
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [
            'single_pair',
            [
                BinaryOutputState(index=0, state=10)
            ],
            b'\x03\x00\x4b\x00\xff\xff\x01\x00\x00\x00\x0a'
        ],
        [
            'three_pairs',
            [
                BinaryOutputState(index=10, state=30),
                BinaryOutputState(index=5, state=10),
                BinaryOutputState(index=1, state=255)
            ],
            b'\x03\x00\x4b\x00\xff\xff\x22\x04\x00\x00\xff\x0a\x1e'
        ],
    ])
    def test_set_binouts(self, _, binouts: List[BinaryOutputState], dpa: bytes):
        request = SetOutputRequest(nadr=3, msgid='setOutputTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.binouts = binouts
        self.json['data']['req']['param']['binOuts'] = [bo.to_json() for bo in binouts]
        self.assertEqual(
            request.binouts,
            binouts
        )
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [-1, 0],
        [255, 0],
        [32, 0],
        [1, -1],
        [1, 256],
        [1, 1000],
    ])
    def test_invalid_binary_output_state_member_constructor(self, index: int, state: int):
        with self.assertRaises(RequestParameterInvalidValueError):
            BinaryOutputState(index=index, state=state)

    @parameterized.expand([
        [-1],
        [32],
        [255],
    ])
    def test_invalid_binary_output_index(self, index: int):
        bo_state = BinaryOutputState(index=0, state=1)
        with self.assertRaises(RequestParameterInvalidValueError):
            bo_state.index = index

    @parameterized.expand([
        [-1],
        [256],
        [1000],
    ])
    def test_invalid_binary_output_state(self, state: int):
        bo_state = BinaryOutputState(index=0, state=1)
        with self.assertRaises(RequestParameterInvalidValueError):
            bo_state.state = state

    @parameterized.expand([
        [
            [
                BinaryOutputState(
                    index=random.randint(0, 31),
                    state=random.randint(0, 255),
                )
                for _ in range(33)
            ]
        ]
    ])
    def test_invalid_state_count(self, binouts: List[BinaryOutputState]):
        with self.assertRaises(RequestParameterInvalidValueError):
            SetOutputRequest(nadr=3, binouts=binouts)



"""UART peripheral request messages."""

from .open import OpenRequest
from .close import CloseRequest
from .write_read import WriteReadRequest
from .clear_write_read import ClearWriteReadRequest

__all__ = [
    'OpenRequest',
    'CloseRequest',
    'WriteReadRequest',
    'ClearWriteReadRequest',
]



import random
import unittest
from typing import List
from parameterized import parameterized
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.peripherals.node.requests.validate_bonds import ValidateBondsRequest, NodeValidateBondsParams


class ValidateBondsRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x01\x00\x01\x08\xff\xff\x0a\x00\x00\x00\x00\x0b\x01\x00\x00\x00'
        self.json = {
            'mType': 'iqrfEmbedNode_ValidateBonds',
            'data': {
                'msgId': 'validateBondsTest',
                'req': {
                    'nAdr': 1,
                    'hwpId': 65535,
                    'param': {
                        'nodes': [
                            {
                                'bondAddr': 10,
                                'mid': 0
                            },
                            {
                                'bondAddr': 11,
                                'mid': 1
                            }
                        ]
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        [
            'single_pair',
            [NodeValidateBondsParams(bond_addr=1, mid=2164554855)],
            b'\x01\x00\x01\x08\xff\xff\x01\x67\x7c\x04\x81'
        ],
        [
            'three_pairs',
            [
                NodeValidateBondsParams(bond_addr=1, mid=0),
                NodeValidateBondsParams(bond_addr=2, mid=2164554855),
                NodeValidateBondsParams(bond_addr=3, mid=2164554771),
            ],
            b'\x01\x00\x01\x08\xff\xff\x01\x00\x00\x00\x00\x02\x67\x7c\x04\x81\x03\x13\x7c\x04\x81'
        ]
    ])
    def test_to_dpa(self, _, params, expected):
        request = ValidateBondsRequest(nadr=1, nodes=params)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        [
            'single_pair',
            [NodeValidateBondsParams(bond_addr=1, mid=2164554855)],
        ],
        [
            'three_pairs',
            [
                NodeValidateBondsParams(bond_addr=1, mid=0),
                NodeValidateBondsParams(bond_addr=2, mid=2164554855),
                NodeValidateBondsParams(bond_addr=3, mid=2164554771),
            ],
        ]
    ])
    def test_to_json(self, _, params):
        request = ValidateBondsRequest(nadr=1, nodes=params, msgid='validateBondsTest')
        self.json['data']['req']['param']['nodes'] = [{'bondAddr': node.bond_addr, 'mid': node.mid} for node in params]
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [
            'single_pair',
            [NodeValidateBondsParams(bond_addr=1, mid=2164554855)],
            b'\x01\x00\x01\x08\xff\xff\x01\x67\x7c\x04\x81'
        ],
        [
            'three_pairs',
            [
                NodeValidateBondsParams(bond_addr=1, mid=0),
                NodeValidateBondsParams(bond_addr=2, mid=2164554855),
                NodeValidateBondsParams(bond_addr=3, mid=2164554771),
            ],
            b'\x01\x00\x01\x08\xff\xff\x01\x00\x00\x00\x00\x02\x67\x7c\x04\x81\x03\x13\x7c\x04\x81'
        ]
    ])
    def test_set_nodes(self, _, params, dpa):
        nodes = [NodeValidateBondsParams(bond_addr=10, mid=0), NodeValidateBondsParams(bond_addr=11, mid=1)]
        request = ValidateBondsRequest(nadr=1, nodes=nodes, msgid='validateBondsTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.nodes = params
        self.json['data']['req']['param']['nodes'] = [{'bondAddr': node.bond_addr, 'mid': node.mid} for node in params]
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [1, -1],
        [256, 10],
        [2, 4294967297],
    ])
    def test_invalid_param_members(self, bond_addr, mid):
        with self.assertRaises(RequestParameterInvalidValueError):
            NodeValidateBondsParams(bond_addr=bond_addr, mid=mid)

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_invalid_bond_addr(self, bond_addr: int):
        params = NodeValidateBondsParams(bond_addr=1, mid=0)
        with self.assertRaises(RequestParameterInvalidValueError):
            params.bond_addr = bond_addr

    @parameterized.expand([
        [-1],
        [4294967297]
    ])
    def test_invalid_mid(self, mid: int):
        params = NodeValidateBondsParams(bond_addr=1, mid=0)
        with self.assertRaises(RequestParameterInvalidValueError):
            params.mid = mid

    @parameterized.expand([
        [[NodeValidateBondsParams(bond_addr=random.randint(0, 255), mid=random.randint(0, 0xFFFFFFFF))] * 12]
    ])
    def test_invalid_pair_count(self, params: List[NodeValidateBondsParams]):
        with self.assertRaises(RequestParameterInvalidValueError):
            ValidateBondsRequest(nadr=1, nodes=params)



"""UART Clear Write Read request message."""

from typing import List, Optional, Union
from iqrfpy.enums.commands import UartRequestCommands
from iqrfpy.enums.message_types import UartMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import RequestParameterInvalidValueError
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.utils.common import Common
from iqrfpy.irequest import IRequest

__all__ = ['ClearWriteReadRequest']


class ClearWriteReadRequest(IRequest):
    """UART Clear Write Read request class."""

    __slots__ = '_read_timeout', '_data'

    def __init__(self, nadr: int, read_timeout: int, data: Optional[List[int]] = None,
                 hwpid: int = dpa_constants.HWPID_MAX, dpa_rsp_time: Optional[float] = None,
                 dev_process_time: Optional[float] = None, msgid: Optional[str] = None):
        """Clear Write Read request constructor.

        Args:
            nadr (int): Device address.
            read_timeout (int): Timeout in 10ms unit to wait for data to be read. To read no data, use value 255.
            data (List[int], optional): Data to write to TX buffer.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate(read_timeout=read_timeout, data=data)
        super().__init__(
            nadr=nadr,
            pnum=EmbedPeripherals.UART,
            pcmd=UartRequestCommands.CLEAR_WRITE_READ,
            m_type=UartMessages.CLEAR_WRITE_READ,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._read_timeout = read_timeout
        self._data = data if data is not None else []

    def _validate(self, read_timeout: int, data: Optional[List[int]] = None):
        """Validate request parameters.

        Args:
            read_timeout (int): Timeout in 10ms unit to wait for data to be read. To read no data, use value 255.
            data (List[int], optional): Data to write to TX buffer.
        """
        self._validate_read_timeout(read_timeout=read_timeout)
        self._validate_data(data=data)

    @staticmethod
    def _validate_read_timeout(read_timeout: int):
        """Validate read timeout parameter.

        Args:
            read_timeout (int): Timeout in 10ms unit to wait for data to be read. To read no data, use value 255.
        Raises:
            RequestParameterInvalidValueError: If read_timeout is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= read_timeout <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('Read timeout value should be between 0 and 255.')

    @property
    def read_timeout(self) -> int:
        """:obj:`int`: Timeout in 10ms unit to wait for data to be read. To read no data, use value 255.

        Getter and setter.
        """
        return self._read_timeout

    @read_timeout.setter
    def read_timeout(self, value: int):
        self._validate_read_timeout(value)
        self._read_timeout = value

    @staticmethod
    def _validate_data(data: Optional[List[int]] = None):
        """Validate data parameter.

        Args:
            data (List[int], optional): Data to write to TX buffer.
        Raises:
            RequestParameterInvalidValueError: If data contains more than 57 values or if values are not
                in range from 0 to 255.
        """
        if data is None:
            return
        if len(data) > 57:
            raise RequestParameterInvalidValueError('Maximum data length is 57 bytes.')
        if not Common.values_in_byte_range(data):
            raise RequestParameterInvalidValueError('Write data values should be between 0 and 255.')

    @property
    def data(self) -> Optional[List[int]]:
        """:obj:`list` of :obj:`int` or :obj:`None`: Data to write to TX buffer.

        Getter and setter.
        """
        return self._data

    @data.setter
    def data(self, value: Optional[List[int]]):
        self._validate_data(value)
        self._data = value if value is not None else []

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        self._pdata = [self._read_timeout] + self._data
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        self._params = {
            'readTimeout': self._read_timeout,
            'writtenData': self._data,
        }
        return super().to_json()



"""RAM peripheral response messages."""

from .read import ReadResponse
from .write import WriteResponse
from .read_any import ReadAnyResponse

__all__ = [
    'ReadResponse',
    'WriteResponse',
    'ReadAnyResponse'
]



"""Coordinator peripheral response messages."""

from .addr_info import AddrInfoResponse
from .authorize_bond import AuthorizeBondResponse
from .backup import BackupResponse
from .bond_node import BondNodeResponse
from .bonded_devices import BondedDevicesResponse
from .clear_all_bonds import ClearAllBondsResponse
from .discovered_devices import DiscoveredDevicesResponse
from .discovery import DiscoveryResponse
from .remove_bond import RemoveBondResponse
from .restore import RestoreResponse
from .set_dpa_params import SetDpaParamsResponse
from .set_hops import SetHopsResponse
from .set_mid import SetMidResponse
from .smart_connect import SmartConnectResponse

__all__ = [
    'AddrInfoResponse',
    'AuthorizeBondResponse',
    'BackupResponse',
    'BondNodeResponse',
    'BondedDevicesResponse',
    'ClearAllBondsResponse',
    'DiscoveredDevicesResponse',
    'DiscoveryResponse',
    'RemoveBondResponse',
    'RestoreResponse',
    'SetDpaParamsResponse',
    'SetHopsResponse',
    'SetMidResponse',
    'SmartConnectResponse',
]



"""OS Set Security request message."""

from typing import List, Optional, Union
from iqrfpy.enums.commands import OSRequestCommands
from iqrfpy.enums.message_types import OSMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import RequestParameterInvalidValueError
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.utils.common import Common
from iqrfpy.utils.enums import IntEnumMember
from iqrfpy.irequest import IRequest

__all__ = [
    'SetSecurityRequest',
    'OsSecurityType'
]


class OsSecurityType(IntEnumMember):
    """OS Security type enum."""

    ACCESS_PASSWORD = 0
    USER_KEY = 1


class SetSecurityRequest(IRequest):
    """OS Set Security request class."""

    __slots__ = '_security_type', '_data'

    def __init__(self, nadr: int, security_type: Union[OsSecurityType, int], data: List[int],
                 hwpid: int = dpa_constants.HWPID_MAX, dpa_rsp_time: Optional[float] = None,
                 dev_process_time: Optional[float] = None, msgid: Optional[str] = None):
        """Set Security request constructor.

        Args:
            nadr (int): Device address.
            security_type (Union[OsSecurityType, int]): Security type.
            data (List[int]): Security data.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate(security_type, data)
        super().__init__(
            nadr=nadr,
            pnum=EmbedPeripherals.OS,
            pcmd=OSRequestCommands.SET_SECURITY,
            m_type=OSMessages.SET_SECURITY,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._security_type = security_type
        self._data = data

    def _validate(self, security_type: Union[OsSecurityType, int], data: List[int]):
        """Validate request parameter.

        Args:
            security_type (Union[OsSecurityType, int]): Security type.
            data (List[int]): Security data.
        """
        self._validate_security_type(security_type)
        self._validate_data(data)

    @staticmethod
    def _validate_security_type(security_type: Union[OsSecurityType, int]):
        """Validate security type parameter.

        Args:
            security_type (Union[OsSecurityType, int]): Security type.
        Raises:
            RequestParameterInvalidValueError: If security_type is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= security_type <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('Security type value should be between 0 and 255.')

    @property
    def security_type(self) -> Union[OsSecurityType, int]:
        """:obj:`OsSecurityType` or :obj:`int`: Security type.

        Getter and setter.
        """
        return self._security_type

    @security_type.setter
    def security_type(self, value: Union[OsSecurityType, int]):
        self._validate_security_type(value)
        self._security_type = value

    @staticmethod
    def _validate_data(data: List[int]):
        """Validate security data parameter.

        Args:
            data (List[int]): Security data.
        Raises:
            RequestParameterInvalidValueError: If data does not contain 16 values or if values are not
                in range from 0 to 255.
        """
        if len(data) != 16:
            raise RequestParameterInvalidValueError('Data should be a list of 16 values.')
        if not Common.values_in_byte_range(data):
            raise RequestParameterInvalidValueError('Data values should be between 0 and 255.')

    @property
    def data(self) -> List[int]:
        """:obj:`list` of :obj:`int`: Security data.

        Getter and setter.
        """
        return self._data

    @data.setter
    def data(self, value: List[int]):
        self._validate_data(value)
        self._data = value

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        self._pdata = [self._security_type] + self._data
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        self._params = {'type': self._security_type, 'data': self._data}
        return super().to_json()



"""Coordinator Restore request message."""

from typing import List, Optional, Union
from iqrfpy.enums.commands import CoordinatorRequestCommands
from iqrfpy.enums.message_types import CoordinatorMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.utils.common import Common
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.irequest import IRequest

__all__ = ['RestoreRequest']


class RestoreRequest(IRequest):
    """Coordinator Restore request class."""

    __slots__ = ('_network_data',)

    def __init__(self, network_data: List[int], hwpid: int = dpa_constants.HWPID_MAX,
                 dpa_rsp_time: Optional[float] = None, dev_process_time: Optional[float] = None,
                 msgid: Optional[str] = None):
        """Restore request constructor.

        Args:
            network_data (List[int]): Block of Coordinator network info data obtained by Backup request.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate_network_data(network_data)
        super().__init__(
            nadr=dpa_constants.COORDINATOR_NADR,
            pnum=EmbedPeripherals.COORDINATOR,
            pcmd=CoordinatorRequestCommands.RESTORE,
            m_type=CoordinatorMessages.RESTORE,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._network_data = network_data

    @staticmethod
    def _validate_network_data(network_data: List[int]):
        """Validates network data parameter.

        Args:
            network_data (List[int]): Block of Coordinator network info data obtained by Backup request.
        Raises:
            RequestParameterInvalidValueError: If network_data is longer than 49 bytes (values)
                or the values are not between 0 and 255.
        """
        if len(network_data) > dpa_constants.BACKUP_DATA_BLOCK_MAX_LEN:
            raise RequestParameterInvalidValueError('Network data should be at most 49 bytes long.')
        if not Common.values_in_byte_range(network_data):
            raise RequestParameterInvalidValueError('Network data block values should be between 0 and 255.')

    @property
    def network_data(self) -> List[int]:
        """:obj:`list` of :obj:`int`: Block of Coordinator network info data obtained by Backup request.

        Getter and setter.
        """
        return self._network_data

    @network_data.setter
    def network_data(self, value: List[int]) -> None:
        self._validate_network_data(value)
        self._network_data = value

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        self._pdata = self._network_data
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        self._params = {'networkData': self._network_data}
        return super().to_json()



"""FRC peripheral response messages."""

from .send import SendResponse
from .extra_result import ExtraResultResponse
from .send_selective import SendSelectiveResponse
from .set_frc_params import SetFrcParamsResponse

__all__ = [
    'SendResponse',
    'ExtraResultResponse',
    'SendSelectiveResponse',
    'SetFrcParamsResponse',
]



"""IO Triplet parameters."""

from typing import Union
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.utils.dpa import BYTE_MIN, BYTE_MAX, IoConstants

__all__ = [
    'IoTriplet'
]


class IoTriplet:
    """IO Triplet parameter class."""

    __slots__ = '_port', '_mask', '_value'

    def __init__(self, port: Union[IoConstants, int], mask: int, value: int):
        """IO Triplet constructor.

        Args:
            port (Union[IoConstants, int]): Port number.
            mask (int): Port pin mask.
            value (int): Value to write.
        """
        self._validate(port=port, mask=mask, value=value)
        self._port = port
        self._mask = mask
        self._value = value

    def _validate(self, port: int, mask: int, value: int):
        """Validate IO triplet parameters.

        Args:
            port (int): Port number.
            mask (int): Port pin mask.
            value (int): Value to write.
        """
        self._validate_port(port=port)
        self._validate_mask(mask=mask)
        self._validate_value(value=value)

    @staticmethod
    def _validate_port(port: Union[IoConstants, int]):
        """Validate port number.

        Args:
            port (Union[IoConstants, int]): Port number.
        Raises:
            RequestParameterInvalidValueError: If port is less than 0 or greater than 255.
        """
        if not BYTE_MIN <= port <= BYTE_MAX:
            raise RequestParameterInvalidValueError('Port should be between 0 and 255.')

    @property
    def port(self) -> Union[IoConstants, int]:
        """:obj:`int` or :obj:`IoConstants`: Port number.

        Getter and setter.
        """
        return self._port

    @port.setter
    def port(self, val: Union[IoConstants, int]):
        self._validate_port(port=val)
        self._value = val

    @staticmethod
    def _validate_mask(mask: int):
        """Validate port pin mask.

        Args:
            mask (int): Port pin mask.
        Raises:
            RequestParameterInvalidValueError: If mask is less than 0 or greater than 255.
        """
        if not BYTE_MIN <= mask <= BYTE_MAX:
            raise RequestParameterInvalidValueError('Mask should be between 0 and 255.')

    @property
    def mask(self) -> int:
        """:obj:`int`: Port pin mask.

        Getter and setter.
        """
        return self._mask

    @mask.setter
    def mask(self, val: int):
        self._validate_mask(mask=val)
        self._mask = val

    @staticmethod
    def _validate_value(value: int):
        """Validate value to write.

        Args:
            value (int): Value to write.
        Raises:
            RequestParameterInvalidValueError: If value is less than 0 or greater than 255.
        """
        if not BYTE_MIN <= value <= BYTE_MAX:
            raise RequestParameterInvalidValueError('Value should be between 0 and 255.')

    @property
    def value(self) -> int:
        """:obj:`int`: Value to write.

        Getter and setter.
        """
        return self._value

    @value.setter
    def value(self, val: int):
        self._validate_value(value=val)
        self._value = val



import unittest
from parameterized import parameterized
from iqrfpy.enums.commands import ExplorationRequestCommands
from iqrfpy.enums.message_types import GenericMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.async_response import AsyncResponse

data_ok: dict = {
    'msgid': 'async',
    'nadr': 0,
    'hwpid': 0,
    'rcode': 128,
    'dpa_value': 0,
    'dpa': b'\x00\x00\xff\x3f\x00\x00\x80\x00\x17\x04\x00\xfd\x26\x00\x00\x00\x00\x00\x00\x05',
    'json': {
        'mType': 'iqrfRaw',
        'data': {
            'msgId': 'async',
            'rsp': {
                'rData': '00.00.ff.3f.00.00.80.00.17.04.00.fd.26.00.00.00.00.00.00.05'
            },
            'status': 0,
            'insId': 'iqrfgd2-default'
        }
    }
}


class AsyncResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, AsyncResponse.from_dpa(data_ok['dpa']), False],
        ['from_json', data_ok, AsyncResponse.from_json(data_ok['json']), True],
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, EmbedPeripherals.EXPLORATION)
        with self.subTest():
            self.assertEqual(response.pcmd, ExplorationRequestCommands.PERIPHERALS_ENUMERATION_INFORMATION)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        with self.subTest():
            self.assertEqual(response.pdata, list(response_data['dpa']))
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, GenericMessages.RAW)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    def test_from_dpa_invalid(self):
        with self.assertRaises(ValueError):
            AsyncResponse.from_dpa(b'\x00\x00\x00\x80\x00\x00\x00')



"""OS peripheral request messages."""

from .read import ReadRequest
from .reset import ResetRequest
from .restart import RestartRequest
from .read_tr_conf import ReadTrConfRequest
from .write_tr_conf import WriteTrConfRequest, OsTrConfData
from .rfpgm import RfpgmRequest
from .sleep_params import OsSleepParams
from .sleep import SleepRequest
from .set_security import SetSecurityRequest, OsSecurityType
from .batch import BatchRequest
from .batch_data import BatchData

__all__ = [
    'ReadRequest',
    'ResetRequest',
    'RestartRequest',
    'OsTrConfData',
    'ReadTrConfRequest',
    'WriteTrConfRequest',
    'RfpgmRequest',
    'OsSleepParams',
    'SleepRequest',
    'OsSecurityType',
    'SetSecurityRequest',
    'BatchData',
    'BatchRequest'
]



"""Coordinator Remove Bond request message."""

from typing import Optional, Union
from iqrfpy.enums.commands import CoordinatorRequestCommands
from iqrfpy.enums.message_types import CoordinatorMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import RequestParameterInvalidValueError
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.irequest import IRequest

__all__ = ['RemoveBondRequest']


class RemoveBondRequest(IRequest):
    """Coordinator RemoveBond request class."""

    __slots__ = ('_bond_addr',)

    def __init__(self, bond_addr: int, hwpid: int = dpa_constants.HWPID_MAX, dpa_rsp_time: Optional[float] = None,
                 dev_process_time: Optional[float] = None, msgid: Optional[str] = None):
        """RemoveBond request constructor.

        Args:
            bond_addr (int): Address of node device to remove from list of bonded nodes.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate(bond_addr)
        super().__init__(
            nadr=dpa_constants.COORDINATOR_NADR,
            pnum=EmbedPeripherals.COORDINATOR,
            pcmd=CoordinatorRequestCommands.REMOVE_BOND,
            m_type=CoordinatorMessages.REMOVE_BOND,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._bond_addr = bond_addr

    @staticmethod
    def _validate(bond_addr: int) -> None:
        """Validates removed node address parameter.

        Args:
            bond_addr (int): Address of node device to remove from list of bonded nodes.
        Raises:
            RequestParameterInvalidValueError: If bond_addr is less than 0 or greater than 255.
        """
        if bond_addr < dpa_constants.BYTE_MIN or bond_addr > dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('Bond address value should be between 0 and 255.')

    @property
    def bond_addr(self) -> int:
        """:obj:`int`: Address of node device to remove from list of bonded nodes.

        Getter and setter.
        """
        return self._bond_addr

    @bond_addr.setter
    def bond_addr(self, value: int) -> None:
        self._validate(value)
        self._bond_addr = value

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        self._pdata = [self._bond_addr]
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        self._params = {'bondAddr': self._bond_addr}
        return super().to_json()
